@@ -2,19 +2,22 @@ package com.artifex.mupdf.fitz;
 
 public class DisplayList
 {
+	static {
+		Context.init();
+	}
+
 	private long pointer;
 
 	protected native void finalize();
 
 	public void destroy() {
 		finalize();
-		pointer = 0;
 	}
 
-	private native long newNative();
+	private native long newNative(Rect mediabox);
 
-	public DisplayList() {
-		pointer = newNative();
+	public DisplayList(Rect mediabox) {
+		pointer = newNative(mediabox);
 	}
 
 	private DisplayList(long p) {
@@ -28,7 +31,7 @@ public class DisplayList
 		return toStructuredText(null);
 	}
 
-	public native Rect[] search(String needle);
+	public native Quad[] search(String needle);
 
 	public native void run(Device dev, Matrix ctm, Rect scissor, Cookie cookie);
 
