@@ -12,7 +12,6 @@ public class DocumentWriter
 
 	public void destroy() {
 		finalize();
-		pointer = 0;
 	}
 
 	private native long newNativeDocumentWriter(String filename, String format, String options);
@@ -21,7 +20,7 @@ public class DocumentWriter
 		pointer = newNativeDocumentWriter(filename, format, options);
 	}
 
-	public native Device beingPage(Rect mediabox);
-	public native void endPage(Device device);
+	public native Device beginPage(Rect mediabox);
+	public native void endPage();
 	public native void close();
 }
