@@ -2,26 +2,24 @@ package com.artifex.mupdf.fitz;
 
 public class Text implements TextWalker
 {
+	static {
+		Context.init();
+	}
+
 	private long pointer;
 
 	protected native void finalize();
 
 	public void destroy() {
 		finalize();
-		pointer = 0;
 	}
 
 	private native long newNative();
-	private native long cloneNative(Text old);
 
 	private Text(long p) {
 		pointer = p;
 	}
 
-	public Text(Text old) {
-		pointer = cloneNative(old);
-	}
-
 	public Text() {
 		pointer = newNative();
 	}
