@@ -1,5 +1,7 @@
 #include "mupdf/fitz.h"
 
+#include <string.h>
+
 static const unsigned char pkm[256*8] =
 {
 	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
@@ -265,18 +267,28 @@ fz_new_bitmap(fz_context *ctx, int w, int h, int n, int xres, int yres)
 {
 	fz_bitmap *bit;
 
-	bit = fz_malloc_struct(ctx, fz_bitmap);
-	bit->refs = 1;
-	bit->w = w;
-	bit->h = h;
-	bit->n = n;
-	bit->xres = xres;
-	bit->yres = yres;
-	/* Span is 32 bit aligned. We may want to make this 64 bit if we
-	 * use SSE2 etc. */
-	bit->stride = ((n * w + 31) & ~31) >> 3;
+	/* Stride is 32 bit aligned. We may want to make this 64 bit if we use SSE2 etc. */
+	int stride = ((n * w + 31) & ~31) >> 3;
+	if (h < 0 || ((size_t)h > (size_t)(SIZE_MAX / stride)))
+		fz_throw(ctx, FZ_ERROR_MEMORY, "bitmap too large");
 
-	bit->samples = fz_malloc_array(ctx, h, bit->stride);
+	bit = fz_malloc_struct(ctx, fz_bitmap);
+	fz_try(ctx)
+	{
+		bit->refs = 1;
+		bit->w = w;
+		bit->h = h;
+		bit->n = n;
+		bit->xres = xres;
+		bit->yres = yres;
+		bit->stride = stride;
+		bit->samples = Memento_label(fz_malloc(ctx, (size_t)h * bit->stride), "bitmap_samples");
+	}
+	fz_catch(ctx)
+	{
+		fz_free(ctx, bit);
+		fz_rethrow(ctx);
+	}
 
 	return bit;
 }
@@ -300,26 +312,32 @@ fz_drop_bitmap(fz_context *ctx, fz_bitmap *bit)
 void
 fz_clear_bitmap(fz_context *ctx, fz_bitmap *bit)
 {
-	memset(bit->samples, 0, bit->stride * bit->h);
+	memset(bit->samples, 0, (size_t)bit->stride * bit->h);
 }
 
 static void
-pbm_write_header(fz_context *ctx, fz_band_writer *writer)
+pbm_write_header(fz_context *ctx, fz_band_writer *writer, fz_colorspace *cs)
 {
 	fz_output *out = writer->out;
 	int w = writer->w;
 	int h = writer->h;
 
+	if (writer->s != 0)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "pbms cannot contain spot colors");
+
 	fz_write_printf(ctx, out, "P4\n%d %d\n", w, h);
 }
 
 static void
-pkm_write_header(fz_context *ctx, fz_band_writer *writer)
+pkm_write_header(fz_context *ctx, fz_band_writer *writer, fz_colorspace *cs)
 {
 	fz_output *out = writer->out;
 	int w = writer->w;
 	int h = writer->h;
 
+	if (writer->s != 0)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "pkms cannot contain spot colors");
+
 	fz_write_printf(ctx, out, "P7\nWIDTH %d\nHEIGHT %d\nDEPTH 4\nMAXVAL 255\nTUPLTYPE CMYK\nENDHDR\n", w, h);
 }
 
@@ -334,7 +352,7 @@ fz_write_bitmap_as_pbm(fz_context *ctx, fz_output *out, fz_bitmap *bitmap)
 	writer = fz_new_pbm_band_writer(ctx, out);
 	fz_try(ctx)
 	{
-		fz_write_header(ctx, writer, bitmap->w, bitmap->h, 1, 0, 0, 0, 0);
+		fz_write_header(ctx, writer, bitmap->w, bitmap->h, 1, 0, 0, 0, 0, NULL, NULL);
 		fz_write_band(ctx, writer, bitmap->stride, bitmap->h, bitmap->samples);
 	}
 	fz_always(ctx)
@@ -354,7 +372,7 @@ fz_write_bitmap_as_pkm(fz_context *ctx, fz_output *out, fz_bitmap *bitmap)
 	writer = fz_new_pkm_band_writer(ctx, out);
 	fz_try(ctx)
 	{
-		fz_write_header(ctx, writer, bitmap->w, bitmap->h, 4, 0, 0, 0, 0);
+		fz_write_header(ctx, writer, bitmap->w, bitmap->h, 4, 0, 0, 0, 0, NULL, NULL);
 		fz_write_band(ctx, writer, bitmap->stride, bitmap->h, bitmap->samples);
 	}
 	fz_always(ctx)
@@ -445,7 +463,10 @@ fz_save_bitmap_as_pbm(fz_context *ctx, fz_bitmap *bitmap, const char *filename)
 {
 	fz_output *out = fz_new_output_with_path(ctx, filename, 0);
 	fz_try(ctx)
+	{
 		fz_write_bitmap_as_pbm(ctx, out, bitmap);
+		fz_close_output(ctx, out);
+	}
 	fz_always(ctx)
 		fz_drop_output(ctx, out);
 	fz_catch(ctx)
@@ -457,7 +478,10 @@ fz_save_bitmap_as_pkm(fz_context *ctx, fz_bitmap *bitmap, const char *filename)
 {
 	fz_output *out = fz_new_output_with_path(ctx, filename, 0);
 	fz_try(ctx)
+	{
 		fz_write_bitmap_as_pkm(ctx, out, bitmap);
+		fz_close_output(ctx, out);
+	}
 	fz_always(ctx)
 		fz_drop_output(ctx, out);
 	fz_catch(ctx)
