@@ -1,14 +1,12 @@
 #include "mupdf/fitz.h"
-#include "fitz-imp.h"
+
+#include "pixmap-imp.h"
 
 #include <assert.h>
 #include <string.h>
 
 #if FZ_ENABLE_JPX
 
-typedef struct fz_jpxd_s fz_jpxd;
-typedef struct stream_block_s stream_block;
-
 static void
 jpx_ycc_to_rgb(fz_context *ctx, fz_pixmap *pix, int cbsign, int crsign)
 {
@@ -56,7 +54,7 @@ jpx_ycc_to_rgb(fz_context *ctx, fz_pixmap *pix, int cbsign, int crsign)
 	(cs) == cJP2_Colorspace_Palette_ICCa || \
 	(cs) == cJP2_Colorspace_Palette_CMYKa)
 
-struct fz_jpxd_s
+typedef struct
 {
 	fz_pixmap *pix;
 	JP2_Palette_Params *palette;
@@ -70,19 +68,19 @@ struct fz_jpxd_s
 	JP2_Property_Value vstep[MAX_COMPONENTS];
 	JP2_Property_Value bpss[MAX_COMPONENTS];
 	JP2_Property_Value signs[MAX_COMPONENTS];
-};
+} fz_jpxd;
 
-struct stream_block_s
+typedef struct
 {
 	const unsigned char *data;
 	size_t size;
-};
+} stream_block;
 
 static void * JP2_Callback_Conv
 jpx_alloc(long size, JP2_Callback_Param param)
 {
 	fz_context *ctx = (fz_context *) param;
-	return fz_malloc(ctx, size);
+	return Memento_label(fz_malloc(ctx, size), "jpx_alloc");
 }
 
 static JP2_Error JP2_Callback_Conv
@@ -115,9 +113,8 @@ jpx_write(unsigned char * pucData, short sComponent, unsigned long ulRow,
 	fz_jpxd *state = (fz_jpxd *) param;
 	JP2_Property_Value hstep, vstep;
 	unsigned char *row;
-	int w, h, n, entries, expand;
+	int w, h, n, k, entries, expand;
 	JP2_Property_Value x, y, i, bps, sign;
-	JP2_Property_Value k;
 	unsigned long **palette;
 
 	w = state->pix->w;
@@ -302,7 +299,7 @@ jpx_read_image(fz_context *ctx, fz_jpxd *state, const unsigned char *data, size_
 
 		if (defcs)
 		{
-			if (defcs->n == nchans)
+			if ((JP2_Property_Value)defcs->n == nchans)
 				state->cs = fz_keep_colorspace(ctx, defcs);
 			else
 				fz_warn(ctx, "jpx file (%lu) and dict colorspace (%d, %s) do not match", nchans, defcs->n, defcs->name);
@@ -330,7 +327,7 @@ jpx_read_image(fz_context *ctx, fz_jpxd *state, const unsigned char *data, size_
 			fz_catch(ctx)
 				fz_warn(ctx, "ignoring embedded ICC profile in JPX");
 
-			if (state->cs && state->cs->n != n)
+			if (state->cs && (JP2_Property_Value)state->cs->n != nchans)
 			{
 				fz_warn(ctx, "invalid number of components in ICC profile, ignoring ICC profile in JPX");
 				fz_drop_colorspace(ctx, state->cs);
@@ -486,21 +483,21 @@ fz_load_jpx_info(fz_context *ctx, const unsigned char *data, size_t size, int *w
 
 #include <openjpeg.h>
 
-struct fz_jpxd_s
+typedef struct
 {
 	int width;
 	int height;
 	fz_colorspace *cs;
 	int xres;
 	int yres;
-};
+} fz_jpxd;
 
-struct stream_block_s
+typedef struct
 {
 	const unsigned char *data;
 	OPJ_SIZE_T size;
 	OPJ_SIZE_T pos;
-};
+} stream_block;
 
 /* OpenJPEG does not provide a safe mechanism to intercept
  * allocations. In the latest version all allocations go
@@ -555,7 +552,7 @@ void *opj_malloc(size_t size)
 
 	assert(ctx != NULL);
 
-	return fz_malloc_no_throw(ctx, size);
+	return Memento_label(fz_malloc_no_throw(ctx, size), "opj_malloc");
 }
 
 void *opj_calloc(size_t n, size_t size)
@@ -588,7 +585,7 @@ void opj_free(void *ptr)
 static void * opj_aligned_malloc_n(size_t alignment, size_t size)
 {
 	uint8_t *ptr;
-	int off;
+	size_t off;
 
 	if (size == 0)
 		return NULL;
@@ -598,7 +595,7 @@ static void * opj_aligned_malloc_n(size_t alignment, size_t size)
 	if (ptr == NULL)
 		return NULL;
 	off = alignment-(((int)(intptr_t)ptr) & (alignment - 1));
-	ptr[off-1] = off;
+	ptr[off-1] = (uint8_t)off;
 	return ptr + off;
 }
 
@@ -635,7 +632,7 @@ static void fz_opj_error_callback(const char *msg, void *client_data)
 {
 	fz_context *ctx = (fz_context *)client_data;
 	char buf[200];
-	int n;
+	size_t n;
 	fz_strlcpy(buf, msg, sizeof buf);
 	n = strlen(buf);
 	if (buf[n-1] == '\n')
@@ -647,7 +644,7 @@ static void fz_opj_warning_callback(const char *msg, void *client_data)
 {
 	fz_context *ctx = (fz_context *)client_data;
 	char buf[200];
-	int n;
+	size_t n;
 	fz_strlcpy(buf, msg, sizeof buf);
 	n = strlen(buf);
 	if (buf[n-1] == '\n')
@@ -695,6 +692,143 @@ static OPJ_BOOL fz_opj_stream_seek(OPJ_OFF_T seek_pos, void * p_user_data)
 	return OPJ_TRUE;
 }
 
+static int32_t
+safe_mul32(fz_context *ctx, int32_t a, int32_t b)
+{
+	int64_t res = ((int64_t)a) * b;
+	int32_t res32 = (int32_t)res;
+
+	if ((res32) != res)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "Overflow while decoding jpx");
+	return res32;
+}
+
+static int32_t
+safe_mla32(fz_context *ctx, int32_t a, int32_t b, int32_t c)
+{
+	int64_t res = ((int64_t)a) * b + c;
+	int32_t res32 = (int32_t)res;
+
+	if ((res32) != res)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "Overflow while decoding jpx");
+	return res32;
+}
+
+static inline void
+template_copy_comp(unsigned char *dst0, int w, int h, int stride, const OPJ_INT32 *src, int32_t ox, int32_t oy, OPJ_UINT32 cdx, OPJ_UINT32 cdy, OPJ_UINT32 cw, OPJ_UINT32 ch, OPJ_UINT32 sgnd, OPJ_UINT32 prec, int comps)
+{
+	int x, y;
+
+	for (y = ch; oy + cdy <= 0 && y > 0; y--)
+	{
+		oy += cdy;
+		dst0 += cdy * stride;
+		src += cw;
+	}
+	for (; y > 0; y--)
+	{
+		int32_t dymin = oy;
+		int32_t dywid = cdy;
+		unsigned char *dst1 = dst0 + ox * comps;
+		int32_t oox = ox;
+		const OPJ_INT32 *src0 = src;
+
+		if (dymin < 0)
+			dywid += dymin, dst1 -= dymin * stride, dymin = 0;
+		if (dymin >= h)
+			break;
+		if (dymin + dywid > h)
+			dywid = h - dymin;
+
+		for (x = cw; oox + cdx <= 0 && x > 0; x--)
+		{
+			oox += cdx;
+			dst1 += cdx * comps;
+			src0++;
+		}
+		for (; x > 0; x--)
+		{
+			OPJ_INT32 v;
+			int32_t xx, yy;
+			int32_t dxmin = oox;
+			int32_t dxwid = cdx;
+			unsigned char *dst2;
+
+			v = *src0++;
+
+			if (sgnd)
+				v = v + (1 << (prec - 1));
+			if (prec > 8)
+				v = v >> (prec - 8);
+			else if (prec < 8)
+				v = v << (8 - prec);
+
+			if (dxmin < 0)
+				dxwid += dxmin, dst1 -= dxmin * comps, dxmin = 0;
+			if (dxmin >= w)
+				break;
+			if (dxmin + dxwid > w)
+				dxwid = w - dxmin;
+
+			dst2 = dst1;
+			for (yy = dywid; yy > 0; yy--)
+			{
+				unsigned char *dst3 = dst2;
+				for (xx = dxwid; xx > 0; xx--)
+				{
+					*dst3 = v;
+					dst3 += comps;
+				}
+				dst2 += stride;
+			}
+			dst1 += comps * cdx;
+			oox += cdx;
+		}
+		dst0 += stride * cdy;
+		src += cw;
+		oy += cdy;
+	}
+}
+
+static void
+copy_jpx_to_pixmap(fz_context *ctx, fz_pixmap *img, opj_image_t *jpx)
+{
+	unsigned char *dst;
+	int stride, comps;
+	int w = img->w;
+	int h = img->h;
+	int k;
+
+	stride = fz_pixmap_stride(ctx, img);
+	comps = fz_pixmap_components(ctx, img);
+	dst = fz_pixmap_samples(ctx, img);
+
+	for (k = 0; k < comps; k++)
+	{
+		opj_image_comp_t *comp = &(jpx->comps[k]);
+		OPJ_UINT32 cdx = comp->dx;
+		OPJ_UINT32 cdy = comp->dy;
+		OPJ_UINT32 cw = comp->w;
+		OPJ_UINT32 ch = comp->h;
+		int32_t oy = safe_mul32(ctx, comp->y0, cdy) - jpx->y0;
+		int32_t ox = safe_mul32(ctx, comp->x0, cdx) - jpx->x0;
+		unsigned char *dst0 = dst + oy * stride;
+
+		if (comp->data == NULL)
+			fz_throw(ctx, FZ_ERROR_GENERIC, "No data for JP2 image component %d", k);
+
+		/* Check that none of the following will overflow. */
+		(void)safe_mla32(ctx, ch, cdy, oy);
+		(void)safe_mla32(ctx, cw, cdx, ox);
+
+		if (cdx == 1 && cdy == 1)
+			template_copy_comp(dst0, w, h, stride, comp->data, ox, oy, 1 /*cdx*/, 1 /*cdy*/, cw, ch, comp->sgnd, comp->prec, comps);
+		else
+			template_copy_comp(dst0, w, h, stride, comp->data, ox, oy, cdx, cdy, cw, ch, comp->sgnd, comp->prec, comps);
+		dst++;
+	}
+}
+
 static fz_pixmap *
 jpx_read_image(fz_context *ctx, fz_jpxd *state, const unsigned char *data, size_t size, fz_colorspace *defcs, int onlymeta)
 {
@@ -704,8 +838,8 @@ jpx_read_image(fz_context *ctx, fz_jpxd *state, const unsigned char *data, size_
 	opj_image_t *jpx;
 	opj_stream_t *stream;
 	OPJ_CODEC_FORMAT format;
-	int a, n, w, h;
-	int x, y, k;
+	int a, n, k;
+	int w, h;
 	stream_block sb;
 	OPJ_UINT32 i;
 
@@ -787,11 +921,14 @@ jpx_read_image(fz_context *ctx, fz_jpxd *state, const unsigned char *data, size_
 		}
 	}
 
-	state->width = w = jpx->x1 - jpx->x0;
-	state->height = h = jpx->y1 - jpx->y0;
+	w = state->width = jpx->x1 - jpx->x0;
+	h = state->height = jpx->y1 - jpx->y0;
 	state->xres = 72; /* openjpeg does not read the JPEG 2000 resc box */
 	state->yres = 72; /* openjpeg does not read the JPEG 2000 resc box */
 
+	if (w < 0 || h < 0)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "Unbelievable size for jpx");
+
 	state->cs = NULL;
 
 	if (defcs)
@@ -850,57 +987,10 @@ jpx_read_image(fz_context *ctx, fz_jpxd *state, const unsigned char *data, size_
 
 	fz_try(ctx)
 	{
-		unsigned char *samples;
-		int stride, comps;
-
 		a = !!a; /* ignore any superfluous alpha channels */
 		img = fz_new_pixmap(ctx, state->cs, w, h, NULL, a);
-		stride = fz_pixmap_stride(ctx, img);
-		comps = fz_pixmap_components(ctx, img);
-		samples = fz_pixmap_samples(ctx, img);
-
 		fz_clear_pixmap_with_value(ctx, img, 0);
-
-		for (k = 0; k < comps; k++)
-		{
-			opj_image_comp_t *comp = &(jpx->comps[k]);
-			int oy = comp->y0 * comp->dy - jpx->y0;
-			int ox = comp->x0 * comp->dx - jpx->x0;
-
-			if (comp->data == NULL)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "No data for JP2 image component %d", k);
-
-			for (y = 0; y < comp->h; y++)
-			{
-				for (x = 0; x < comp->w; x++)
-				{
-					OPJ_INT32 v;
-					int dx;
-					int dy;
-
-					v = comp->data[y * comp->w + x];
-
-					if (comp->sgnd)
-						v = v + (1 << (comp->prec - 1));
-					if (comp->prec > 8)
-						v = v >> (comp->prec - 8);
-					else if (comp->prec < 8)
-						v = v << (8 - comp->prec);
-
-					for (dy = 0; dy < comp->dy; dy++)
-					{
-						for (dx = 0; dx < comp->dx; dx++)
-						{
-							int xx = ox + x * comp->dx + dx;
-							int yy = oy + y * comp->dy + dy;
-
-							if (xx < w && yy < h)
-								samples[yy * stride + xx * comps + k] = v;
-						}
-					}
-				}
-			}
-		}
+		copy_jpx_to_pixmap(ctx, img, jpx);
 
 		if (jpx->color_space == OPJ_CLRSPC_SYCC && n == 3 && a == 0)
 			jpx_ycc_to_rgb(ctx, img, 1, 1);
