@@ -1,5 +1,10 @@
 #include "mupdf/fitz.h"
 
+#include "pixmap-imp.h"
+
+#include <string.h>
+#include <limits.h>
+
 enum
 {
 	PAM_UNKNOWN = 0,
@@ -26,6 +31,7 @@ enum
 
 struct info
 {
+	int subimages;
 	fz_colorspace *cs;
 	int width, height;
 	int maxval, bitdepth;
@@ -71,8 +77,8 @@ static inline int bitdepth_from_maxval(int maxval)
 	return depth;
 }
 
-static unsigned char *
-pnm_read_white(fz_context *ctx, unsigned char *p, unsigned char *e, int single_line)
+static const unsigned char *
+pnm_read_white(fz_context *ctx, const unsigned char *p, const unsigned char *e, int single_line)
 {
 	if (e - p < 1)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse whitespace in pnm image");
@@ -116,8 +122,8 @@ pnm_read_white(fz_context *ctx, unsigned char *p, unsigned char *e, int single_l
 	return p;
 }
 
-static unsigned char *
-pnm_read_signature(fz_context *ctx, unsigned char *p, unsigned char *e, char *signature)
+static const unsigned char *
+pnm_read_signature(fz_context *ctx, const unsigned char *p, const unsigned char *e, char *signature)
 {
 	if (e - p < 2)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse magic number in pnm image");
@@ -129,25 +135,26 @@ pnm_read_signature(fz_context *ctx, unsigned char *p, unsigned char *e, char *si
 	return p;
 }
 
-static unsigned char *
-pnm_read_number(fz_context *ctx, unsigned char *p, unsigned char *e, int *number)
+static const unsigned char *
+pnm_read_number(fz_context *ctx, const unsigned char *p, const unsigned char *e, int *number)
 {
 	if (e - p < 1)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse number in pnm image");
-	if (*p < '0' && *p > '9')
+	if (*p < '0' || *p > '9')
 		fz_throw(ctx, FZ_ERROR_GENERIC, "expected numeric field in pnm image");
 
 	while (p < e && *p >= '0' && *p <= '9')
 	{
-		*number = *number * 10 + *p - '0';
+		if (number)
+			*number = *number * 10 + *p - '0';
 		p++;
 	}
 
 	return p;
 }
 
-static unsigned char *
-pnm_read_tupletype(fz_context *ctx, unsigned char *p, unsigned char *e, int *tupletype)
+static const unsigned char *
+pnm_read_tupletype(fz_context *ctx, const unsigned char *p, const unsigned char *e, int *tupletype)
 {
 	const struct { int len; char *str; int type; } tupletypes[] =
 	{
@@ -160,7 +167,7 @@ pnm_read_tupletype(fz_context *ctx, unsigned char *p, unsigned char *e, int *tup
 		{4, "CMYK", PAM_CMYK},
 		{10, "CMYK_ALPHA", PAM_CMYKA},
 	};
-	unsigned char *s;
+	const unsigned char *s;
 	int i, len;
 
 	if (e - p < 1)
@@ -171,7 +178,7 @@ pnm_read_tupletype(fz_context *ctx, unsigned char *p, unsigned char *e, int *tup
 		p++;
 	len = p - s;
 
-	for (i = 0; i < nelem(tupletypes); i++)
+	for (i = 0; i < (int)nelem(tupletypes); i++)
 		if (len == tupletypes[i].len && !strncmp((char *) s, tupletypes[i].str, len))
 		{
 			*tupletype = tupletypes[i].type;
@@ -181,8 +188,8 @@ pnm_read_tupletype(fz_context *ctx, unsigned char *p, unsigned char *e, int *tup
 	fz_throw(ctx, FZ_ERROR_GENERIC, "unknown tuple type in pnm image");
 }
 
-static unsigned char *
-pnm_read_token(fz_context *ctx, unsigned char *p, unsigned char *e, int *token)
+static const unsigned char *
+pnm_read_token(fz_context *ctx, const unsigned char *p, const unsigned char *e, int *token)
 {
 	const struct { int len; char *str; int type; } tokens[] =
 	{
@@ -193,7 +200,7 @@ pnm_read_token(fz_context *ctx, unsigned char *p, unsigned char *e, int *token)
 		{8, "TUPLTYPE", TOKEN_TUPLTYPE},
 		{6, "ENDHDR", TOKEN_ENDHDR},
 	};
-	unsigned char *s;
+	const unsigned char *s;
 	int i, len;
 
 	if (e - p < 1)
@@ -204,7 +211,7 @@ pnm_read_token(fz_context *ctx, unsigned char *p, unsigned char *e, int *token)
 		p++;
 	len = p - s;
 
-	for (i = 0; i < nelem(tokens); i++)
+	for (i = 0; i < (int)nelem(tokens); i++)
 		if (len == tokens[i].len && !strncmp((char *) s, tokens[i].str, len))
 		{
 			*token = tokens[i].type;
@@ -222,7 +229,7 @@ map_color(fz_context *ctx, int color, int inmax, int outmax)
 }
 
 static fz_pixmap *
-pnm_ascii_read_image(fz_context *ctx, struct info *pnm, unsigned char *p, unsigned char *e, int onlymeta, int bitmap)
+pnm_ascii_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p, const unsigned char *e, int onlymeta, int bitmap, const unsigned char **out)
 {
 	fz_pixmap *img = NULL;
 
@@ -255,13 +262,42 @@ pnm_ascii_read_image(fz_context *ctx, struct info *pnm, unsigned char *p, unsign
 	if ((unsigned int)pnm->height > UINT_MAX / pnm->width / fz_colorspace_n(ctx, pnm->cs) / (pnm->bitdepth / 8 + 1))
 		fz_throw(ctx, FZ_ERROR_GENERIC, "image too large");
 
-	if (!onlymeta)
+	if (onlymeta)
+	{
+		int x, y, k;
+		int w, h, n;
+
+		w = pnm->width;
+		h = pnm->height;
+		n = fz_colorspace_n(ctx, pnm->cs);
+
+		if (bitmap)
+		{
+			for (y = 0; y < h; y++)
+				for (x = -1; x < w; x++)
+				{
+					p = pnm_read_number(ctx, p, e, NULL);
+					p = pnm_read_white(ctx, p, e, 0);
+				}
+		}
+		else
+		{
+			for (y = 0; y < h; y++)
+				for (x = 0; x < w; x++)
+					for (k = 0; k < n; k++)
+					{
+						p = pnm_read_number(ctx, p, e, NULL);
+						p = pnm_read_white(ctx, p, e, 0);
+					}
+		}
+	}
+	else
 	{
 		unsigned char *dp;
 		int x, y, k;
 		int w, h, n;
 
-		img = fz_new_pixmap(ctx, pnm->cs, pnm->width, pnm->height, 0);
+		img = fz_new_pixmap(ctx, pnm->cs, pnm->width, pnm->height, NULL, 0);
 		dp = img->samples;
 
 		w = img->w;
@@ -296,27 +332,34 @@ pnm_ascii_read_image(fz_context *ctx, struct info *pnm, unsigned char *p, unsign
 		}
 	}
 
+	if (out)
+		*out = p;
+
 	return img;
 }
 
 static fz_pixmap *
-pnm_binary_read_image(fz_context *ctx, struct info *pnm, unsigned char *p, unsigned char *e, int onlymeta, int bitmap)
+pnm_binary_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p, const unsigned char *e, int onlymeta, int bitmap, const unsigned char **out)
 {
 	fz_pixmap *img = NULL;
 
+	pnm->width = 0;
 	p = pnm_read_number(ctx, p, e, &pnm->width);
 	p = pnm_read_white(ctx, p, e, 0);
 
 	if (bitmap)
 	{
+		pnm->height = 0;
 		p = pnm_read_number(ctx, p, e, &pnm->height);
 		p = pnm_read_white(ctx, p, e, 1);
 		pnm->maxval = 1;
 	}
 	else
 	{
+		pnm->height = 0;
 		p = pnm_read_number(ctx, p, e, &pnm->height);
 		p = pnm_read_white(ctx, p, e, 0);
+		pnm->maxval = 0;
 		p = pnm_read_number(ctx, p, e, &pnm->maxval);
 		p = pnm_read_white(ctx, p, e, 1);
 	}
@@ -333,13 +376,28 @@ pnm_binary_read_image(fz_context *ctx, struct info *pnm, unsigned char *p, unsig
 	if ((unsigned int)pnm->height > UINT_MAX / pnm->width / fz_colorspace_n(ctx, pnm->cs) / (pnm->bitdepth / 8 + 1))
 		fz_throw(ctx, FZ_ERROR_GENERIC, "image too large");
 
-	if (!onlymeta)
+	if (onlymeta)
+	{
+		int w = pnm->width;
+		int h = pnm->height;
+		int n = fz_colorspace_n(ctx, pnm->cs);
+
+		if (pnm->maxval == 255)
+			p += n * w * h;
+		else if (bitmap)
+			p += ((w + 7) / 8) * h;
+		else if (pnm->maxval < 255)
+			p += n * w * h;
+		else
+			p += 2 * n * w * h;
+	}
+	else
 	{
 		unsigned char *dp;
 		int x, y, k;
 		int w, h, n;
 
-		img = fz_new_pixmap(ctx, pnm->cs, pnm->width, pnm->height, 0);
+		img = fz_new_pixmap(ctx, pnm->cs, pnm->width, pnm->height, NULL, 0);
 		dp = img->samples;
 
 		w = img->w;
@@ -347,7 +405,10 @@ pnm_binary_read_image(fz_context *ctx, struct info *pnm, unsigned char *p, unsig
 		n = img->n;
 
 		if (pnm->maxval == 255)
-			memcpy(dp, p, w * h * n);
+		{
+			memcpy(dp, p, (size_t)w * h * n);
+			p += n * w * h;
+		}
 		else if (bitmap)
 		{
 			for (y = 0; y < h; y++)
@@ -381,44 +442,47 @@ pnm_binary_read_image(fz_context *ctx, struct info *pnm, unsigned char *p, unsig
 		}
 	}
 
+	if (out)
+		*out = p;
+
 	return img;
 }
 
-static unsigned char *
-pam_binary_read_header(fz_context *ctx, struct info *pnm, unsigned char *p, unsigned char *e)
+static const unsigned char *
+pam_binary_read_header(fz_context *ctx, struct info *pnm, const unsigned char *p, const unsigned char *e)
 {
 	int token = TOKEN_UNKNOWN;
 
-	fz_try(ctx)
-	{
-		while (p < e && token != TOKEN_ENDHDR)
-		{
-			p = pnm_read_token(ctx, p, e, &token);
-			p = pnm_read_white(ctx, p, e, 0);
-			switch (token)
-			{
-			case TOKEN_WIDTH: p = pnm_read_number(ctx, p, e, &pnm->width); break;
-			case TOKEN_HEIGHT: p = pnm_read_number(ctx, p, e, &pnm->height); break;
-			case TOKEN_DEPTH: p = pnm_read_number(ctx, p, e, &pnm->depth); break;
-			case TOKEN_MAXVAL: p = pnm_read_number(ctx, p, e, &pnm->maxval); break;
-			case TOKEN_TUPLTYPE: p = pnm_read_tupletype(ctx, p, e, &pnm->tupletype); break;
-			case TOKEN_ENDHDR: break;
-			default:
-				   fz_throw(ctx, FZ_ERROR_GENERIC, "unknown header token in pnm image");
-			}
+	pnm->width = 0;
+	pnm->height = 0;
+	pnm->depth = 0;
+	pnm->maxval = 0;
+	pnm->tupletype = 0;
 
-			if (token != TOKEN_ENDHDR)
-				p = pnm_read_white(ctx, p, e, 0);
+	while (p < e && token != TOKEN_ENDHDR)
+	{
+		p = pnm_read_token(ctx, p, e, &token);
+		p = pnm_read_white(ctx, p, e, 0);
+		switch (token)
+		{
+		case TOKEN_WIDTH: p = pnm_read_number(ctx, p, e, &pnm->width); break;
+		case TOKEN_HEIGHT: p = pnm_read_number(ctx, p, e, &pnm->height); break;
+		case TOKEN_DEPTH: p = pnm_read_number(ctx, p, e, &pnm->depth); break;
+		case TOKEN_MAXVAL: p = pnm_read_number(ctx, p, e, &pnm->maxval); break;
+		case TOKEN_TUPLTYPE: p = pnm_read_tupletype(ctx, p, e, &pnm->tupletype); break;
+		case TOKEN_ENDHDR: break;
+		default: fz_throw(ctx, FZ_ERROR_GENERIC, "unknown header token in pnm image");
 		}
+
+		if (token != TOKEN_ENDHDR)
+			p = pnm_read_white(ctx, p, e, 0);
 	}
-	fz_catch(ctx)
-		fz_rethrow(ctx);
 
 	return p;
 }
 
 static fz_pixmap *
-pam_binary_read_image(fz_context *ctx, struct info *pnm, unsigned char *p, unsigned char *e, int onlymeta)
+pam_binary_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p, const unsigned char *e, int onlymeta, const unsigned char **out)
 {
 	fz_pixmap *img = NULL;
 	int bitmap = 0;
@@ -497,13 +561,52 @@ pam_binary_read_image(fz_context *ctx, struct info *pnm, unsigned char *p, unsig
 	if ((unsigned int)pnm->height > UINT_MAX / pnm->width / fz_colorspace_n(ctx, pnm->cs) / (pnm->bitdepth / 8 + 1))
 		fz_throw(ctx, FZ_ERROR_GENERIC, "image too large");
 
+	if (onlymeta)
+	{
+		int packed;
+		int w, h, n;
+		size_t size;
+
+		w = pnm->width;
+		h = pnm->height;
+		n = fz_colorspace_n(ctx, pnm->cs) + pnm->alpha;
+
+		/* some encoders incorrectly pack bits into bytes and invert the image */
+		packed = 0;
+		size = (size_t)w * h * n;
+		if (pnm->maxval == 1)
+		{
+			const unsigned char *e_packed = p + size / 8;
+			if (e_packed < e - 1 && e_packed[0] == 'P' && e_packed[1] >= '0' && e_packed[1] <= '7')
+				e = e_packed;
+			if (e < p || (size_t)(e - p) < size)
+				packed = 1;
+		}
+		if (packed && (e < p || (size_t)(e - p) < size / 8))
+			fz_throw(ctx, FZ_ERROR_GENERIC, "truncated packed image");
+		if (!packed && (e < p || (size_t)(e - p) < size * (pnm->maxval < 256 ? 1 : 2)))
+			fz_throw(ctx, FZ_ERROR_GENERIC, "truncated image");
+
+		if (pnm->maxval == 255)
+			p += size;
+		else if (bitmap && packed)
+			p += ((w + 7) / 8) * h;
+		else if (bitmap)
+			p += size;
+		else if (pnm->maxval < 255)
+			p += size;
+		else
+			p += 2 * size;
+	}
+
 	if (!onlymeta)
 	{
 		unsigned char *dp;
-		int x, y, k, packed = 0;
+		int x, y, k, packed;
 		int w, h, n;
+		size_t size;
 
-		img = fz_new_pixmap(ctx, pnm->cs, pnm->width, pnm->height, pnm->alpha);
+		img = fz_new_pixmap(ctx, pnm->cs, pnm->width, pnm->height, NULL, pnm->alpha);
 		fz_try(ctx)
 		{
 			dp = img->samples;
@@ -512,20 +615,26 @@ pam_binary_read_image(fz_context *ctx, struct info *pnm, unsigned char *p, unsig
 			h = img->h;
 			n = img->n;
 
-			if (pnm->maxval == 1 && e - p < w * h * n)
+			/* some encoders incorrectly pack bits into bytes and invert the image */
+			size = (size_t)w * h * n;
+			packed = 0;
+			if (pnm->maxval == 1)
 			{
-				if (e - p < w * h * n / 8)
-					fz_throw(ctx, FZ_ERROR_GENERIC, "truncated image");
-				packed = 1;
+				const unsigned char *e_packed = p + size / 8;
+				if (e_packed < e - 1 && e_packed[0] == 'P' && e_packed[1] >= '0' && e_packed[1] <= '7')
+					e = e_packed;
+				if (e < p || (size_t)(e - p) < size)
+					packed = 1;
 			}
-			else if (e - p < w * h * n * (pnm->maxval < 256 ? 1 : 2))
+			if (packed && (e < p || (size_t)(e - p) < size / 8))
+				fz_throw(ctx, FZ_ERROR_GENERIC, "truncated packed image");
+			if (!packed && (e < p || (size_t)(e - p) < size * (pnm->maxval < 256 ? 1 : 2)))
 				fz_throw(ctx, FZ_ERROR_GENERIC, "truncated image");
 
 			if (pnm->maxval == 255)
-				memcpy(dp, p, w * h * n);
+				memcpy(dp, p, size);
 			else if (bitmap && packed)
 			{
-				/* some encoders incorrectly pack bits into bytes and inverts the image */
 				for (y = 0; y < h; y++)
 					for (x = 0; x < w; x++)
 					{
@@ -565,10 +674,7 @@ pam_binary_read_image(fz_context *ctx, struct info *pnm, unsigned char *p, unsig
 			}
 
 			if (pnm->alpha)
-			{
-				img = fz_ensure_pixmap_is_additive(ctx, img);
 				fz_premultiply_pixmap(ctx, img);
-			}
 		}
 		fz_catch(ctx)
 		{
@@ -577,86 +683,103 @@ pam_binary_read_image(fz_context *ctx, struct info *pnm, unsigned char *p, unsig
 		}
 	}
 
+	if (out)
+		*out = p;
+
 	return img;
 }
 
 static fz_pixmap *
-pnm_read_image(fz_context *ctx, struct info *pnm, unsigned char *p, size_t total, int onlymeta)
+pnm_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p, size_t total, int onlymeta, int subimage)
 {
-	unsigned char *e = p + total;
+	const unsigned char *e = p + total;
 	char signature[3] = { 0 };
+	fz_pixmap *pix = NULL;
 
-	p = pnm_read_signature(ctx, p, e, signature);
-	p = pnm_read_white(ctx, p, e, 0);
+	while (p < e && ((!onlymeta && subimage >= 0) || onlymeta))
+	{
+		int subonlymeta = onlymeta || (subimage > 0);
 
-	if (!strcmp(signature, "P1"))
-	{
-		pnm->cs = fz_device_gray(ctx);
-		return pnm_ascii_read_image(ctx, pnm, p, e, onlymeta, 1);
+		p = pnm_read_signature(ctx, p, e, signature);
+		p = pnm_read_white(ctx, p, e, 0);
+
+		if (!strcmp(signature, "P1"))
+		{
+			pnm->cs = fz_device_gray(ctx);
+			pix = pnm_ascii_read_image(ctx, pnm, p, e, subonlymeta, 1, &p);
+		}
+		else if (!strcmp(signature, "P2"))
+		{
+			pnm->cs = fz_device_gray(ctx);
+			pix = pnm_ascii_read_image(ctx, pnm, p, e, subonlymeta, 0, &p);
+		}
+		else if (!strcmp(signature, "P3"))
+		{
+			pnm->cs = fz_device_rgb(ctx);
+			pix = pnm_ascii_read_image(ctx, pnm, p, e, subonlymeta, 0, &p);
+		}
+		else if (!strcmp(signature, "P4"))
+		{
+			pnm->cs = fz_device_gray(ctx);
+			pix = pnm_binary_read_image(ctx, pnm, p, e, subonlymeta, 1, &p);
+		}
+		else if (!strcmp(signature, "P5"))
+		{
+			pnm->cs = fz_device_gray(ctx);
+			pix = pnm_binary_read_image(ctx, pnm, p, e, subonlymeta, 0, &p);
+		}
+		else if (!strcmp(signature, "P6"))
+		{
+			pnm->cs = fz_device_rgb(ctx);
+			pix = pnm_binary_read_image(ctx, pnm, p, e, subonlymeta, 0, &p);
+		}
+		else if (!strcmp(signature, "P7"))
+			pix = pam_binary_read_image(ctx, pnm, p, e, subonlymeta, &p);
+		else
+			fz_throw(ctx, FZ_ERROR_GENERIC, "unsupported portable anymap signature (0x%02x, 0x%02x)", signature[0], signature[1]);
+
+		if (onlymeta)
+			pnm->subimages++;
+		if (subimage >= 0)
+			subimage--;
 	}
-	else if (!strcmp(signature, "P2"))
-	{
-		pnm->cs = fz_device_gray(ctx);
-		return pnm_ascii_read_image(ctx, pnm, p, e, onlymeta, 0);
-	}
-	else if (!strcmp(signature, "P3"))
-	{
-		pnm->cs = fz_device_rgb(ctx);
-		return pnm_ascii_read_image(ctx, pnm, p, e, onlymeta, 0);
-	}
-	else if (!strcmp(signature, "P4"))
-	{
-		pnm->cs = fz_device_gray(ctx);
-		return pnm_binary_read_image(ctx, pnm, p, e, onlymeta, 1);
-	}
-	else if (!strcmp(signature, "P5"))
-	{
-		pnm->cs = fz_device_gray(ctx);
-		return pnm_binary_read_image(ctx, pnm, p, e, onlymeta, 0);
-	}
-	else if (!strcmp(signature, "P6"))
-	{
-		pnm->cs = fz_device_rgb(ctx);
-		return pnm_binary_read_image(ctx, pnm, p, e, onlymeta, 0);
-	}
-	else if (!strcmp(signature, "P7"))
-		return pam_binary_read_image(ctx, pnm, p, e, onlymeta);
-	else
-		fz_throw(ctx, FZ_ERROR_GENERIC, "unsupported portable anymap signature (0x%02x, 0x%02x)", signature[0], signature[1]);
+
+	if (p >= e && subimage >= 0)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "subimage count out of range");
+
+	return pix;
 }
 
 fz_pixmap *
-fz_load_pnm(fz_context *ctx, unsigned char *p, size_t total)
+fz_load_pnm(fz_context *ctx, const unsigned char *p, size_t total)
 {
-	fz_pixmap *img;
 	struct info pnm = { 0 };
-
-	fz_try(ctx)
-		img = pnm_read_image(ctx, &pnm, p, total, 0);
-	fz_always(ctx)
-		fz_drop_colorspace(ctx, pnm.cs);
-	fz_catch(ctx)
-		fz_rethrow(ctx);
-
-	return img;
+	return pnm_read_image(ctx, &pnm, p, total, 0, 0);
 }
 
 void
-fz_load_pnm_info(fz_context *ctx, unsigned char *p, size_t total, int *wp, int *hp, int *xresp, int *yresp, fz_colorspace **cspacep)
+fz_load_pnm_info(fz_context *ctx, const unsigned char *p, size_t total, int *wp, int *hp, int *xresp, int *yresp, fz_colorspace **cspacep)
 {
 	struct info pnm = { 0 };
+	(void) pnm_read_image(ctx, &pnm, p, total, 1, 0);
+	*cspacep = fz_keep_colorspace(ctx, pnm.cs); /* pnm.cs is a borrowed device colorspace */
+	*wp = pnm.width;
+	*hp = pnm.height;
+	*xresp = 72;
+	*yresp = 72;
+}
 
-	fz_try(ctx)
-	{
-		pnm_read_image(ctx, &pnm, p, total, 1);
-		*cspacep = pnm.cs;
-		*wp = pnm.width;
-		*hp = pnm.height;
-		*xresp = 72;
-		*yresp = 72;
-	}
-	fz_always(ctx)
-		fz_drop_colorspace(ctx, pnm.cs);
-	fz_catch(ctx)
-		fz_rethrow(ctx);
+fz_pixmap *
+fz_load_pnm_subimage(fz_context *ctx, const unsigned char *p, size_t total, int subimage)
+{
+	struct info pnm = { 0 };
+	return pnm_read_image(ctx, &pnm, p, total, 0, subimage);
+}
+
+int
+fz_load_pnm_subimage_count(fz_context *ctx, const unsigned char *p, size_t total)
+{
+	struct info pnm = { 0 };
+	(void) pnm_read_image(ctx, &pnm, p, total, 1, -1);
+	return pnm.subimages;
 }
