@@ -1,5 +1,8 @@
 #include "mupdf/fitz.h"
 
+#include <string.h>
+#include <assert.h>
+
 // Thoughts for further optimisations:
 // All paths start with MoveTo. We could probably avoid most cases where
 // we store that. The next thing after a close must be a move.
@@ -12,7 +15,7 @@
 // a trailing move. Trailing moves can always be stripped when path
 // construction completes.
 
-typedef enum fz_path_command_e
+typedef enum
 {
 	FZ_MOVETO = 'M',
 	FZ_LINETO = 'L',
@@ -35,7 +38,7 @@ typedef enum fz_path_command_e
 	FZ_QUADTOCLOSE = 'q',
 } fz_path_item_kind;
 
-struct fz_path_s
+struct fz_path
 {
 	int8_t refs;
 	uint8_t packed;
@@ -47,7 +50,7 @@ struct fz_path_s
 	fz_point begin;
 };
 
-typedef struct fz_packed_path_s
+typedef struct
 {
 	int8_t refs;
 	uint8_t packed;
@@ -55,6 +58,19 @@ typedef struct fz_packed_path_s
 	uint8_t cmd_len;
 } fz_packed_path;
 
+/*
+	Paths are created UNPACKED. That means we have a fz_path
+	structure with coords and cmds pointing to malloced blocks.
+
+	After they have been completely constructed, callers may choose
+	to 'pack' them into some target block of memory. If if coord_len
+	and cmd_len are both < 256, then they are PACKED_FLAT into an
+	fz_packed_path with the coords and cmds in the bytes afterwards,
+	all inside the target block. If they cannot be accomodated in
+	that way, then they are PACKED_OPEN, where an fz_path is put
+	into the target block, and cmds and coords remain pointers to
+	allocated blocks.
+*/
 enum
 {
 	FZ_PATH_UNPACKED = 0,
@@ -80,6 +96,14 @@ fz_new_path(fz_context *ctx)
 	return path;
 }
 
+/*
+	Take an additional reference to
+	a path.
+
+	No modifications should be carried out on a path
+	to which more than one reference is held, as
+	this can cause race conditions.
+*/
 fz_path *
 fz_keep_path(fz_context *ctx, const fz_path *pathc)
 {
@@ -130,14 +154,31 @@ int fz_packed_path_size(const fz_path *path)
 	}
 }
 
-int
-fz_pack_path(fz_context *ctx, uint8_t *pack_, int max, const fz_path *path)
+size_t
+fz_pack_path(fz_context *ctx, uint8_t *pack_, size_t max, const fz_path *path)
 {
 	uint8_t *ptr;
-	int size;
+	size_t size;
 
-	if (path->packed)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Can't repack a packed path");
+	if (path->packed == FZ_PATH_PACKED_FLAT)
+	{
+		fz_packed_path *pack = (fz_packed_path *)path;
+		fz_packed_path *out = (fz_packed_path *)pack_;
+		size = sizeof(fz_packed_path) + sizeof(float) * pack->coord_len + sizeof(uint8_t) * pack->cmd_len;
+
+		if (size > max)
+			fz_throw(ctx, FZ_ERROR_GENERIC, "Can't pack a path that small!");
+
+		if (out)
+		{
+			out->refs = 1;
+			out->packed = FZ_PATH_PACKED_FLAT;
+			out->coord_len = pack->coord_len;
+			out->cmd_len = pack->cmd_len;
+			memcpy(&out[1], &pack[1], size - sizeof(*out));
+		}
+		return size;
+	}
 
 	size = sizeof(fz_packed_path) + sizeof(float) * path->coord_len + sizeof(uint8_t) * path->cmd_len;
 
@@ -161,10 +202,10 @@ fz_pack_path(fz_context *ctx, uint8_t *pack_, int max, const fz_path *path)
 			pack->coord_len = path->coord_len;
 			pack->cmd_cap = path->cmd_len;
 			pack->cmd_len = path->cmd_len;
-			pack->coords = fz_malloc_array(ctx, path->coord_len, sizeof(float));
+			pack->coords = Memento_label(fz_malloc_array(ctx, path->coord_len, float), "path_packed_coords");
 			fz_try(ctx)
 			{
-				pack->cmds = fz_malloc_array(ctx, path->cmd_len, sizeof(uint8_t));
+				pack->cmds = Memento_label(fz_malloc_array(ctx, path->cmd_len, uint8_t), "path_packed_cmds");
 			}
 			fz_catch(ctx)
 			{
@@ -205,7 +246,7 @@ push_cmd(fz_context *ctx, fz_path *path, int cmd)
 	if (path->cmd_len + 1 >= path->cmd_cap)
 	{
 		int new_cmd_cap = fz_maxi(16, path->cmd_cap * 2);
-		path->cmds = fz_resize_array(ctx, path->cmds, new_cmd_cap, sizeof(unsigned char));
+		path->cmds = fz_realloc_array(ctx, path->cmds, new_cmd_cap, unsigned char);
 		path->cmd_cap = new_cmd_cap;
 	}
 
@@ -218,7 +259,7 @@ push_coord(fz_context *ctx, fz_path *path, float x, float y)
 	if (path->coord_len + 2 >= path->coord_cap)
 	{
 		int new_coord_cap = fz_maxi(32, path->coord_cap * 2);
-		path->coords = fz_resize_array(ctx, path->coords, new_coord_cap, sizeof(float));
+		path->coords = fz_realloc_array(ctx, path->coords, new_coord_cap, float);
 		path->coord_cap = new_coord_cap;
 	}
 
@@ -235,7 +276,7 @@ push_ord(fz_context *ctx, fz_path *path, float xy, int isx)
 	if (path->coord_len + 1 >= path->coord_cap)
 	{
 		int new_coord_cap = fz_maxi(32, path->coord_cap * 2);
-		path->coords = fz_resize_array(ctx, path->coords, new_coord_cap, sizeof(float));
+		path->coords = fz_realloc_array(ctx, path->coords, new_coord_cap, float);
 		path->coord_cap = new_coord_cap;
 	}
 
@@ -572,13 +613,12 @@ fz_rectto(fz_context *ctx, fz_path *path, float x1, float y1, float x2, float y2
 	path->current = path->begin;
 }
 
-static inline fz_rect *bound_expand(fz_rect *r, const fz_point *p)
+static inline void bound_expand(fz_rect *r, fz_point p)
 {
-	if (p->x < r->x0) r->x0 = p->x;
-	if (p->y < r->y0) r->y0 = p->y;
-	if (p->x > r->x1) r->x1 = p->x;
-	if (p->y > r->y1) r->y1 = p->y;
-	return r;
+	if (p.x < r->x0) r->x0 = p.x;
+	if (p.y < r->y0) r->y0 = p.y;
+	if (p.x > r->x1) r->x1 = p.x;
+	if (p.y > r->y1) r->y1 = p.y;
 }
 
 void fz_walk_path(fz_context *ctx, const fz_path *path, const fz_path_walker *proc, void *arg)
@@ -830,7 +870,7 @@ void fz_walk_path(fz_context *ctx, const fz_path *path, const fz_path_walker *pr
 
 typedef struct
 {
-	const fz_matrix *ctm;
+	fz_matrix ctm;
 	fz_rect rect;
 	fz_point move;
 	int trailing_move;
@@ -841,10 +881,7 @@ static void
 bound_moveto(fz_context *ctx, void *arg_, float x, float y)
 {
 	bound_path_arg *arg = (bound_path_arg *)arg_;
-
-	arg->move.x = x;
-	arg->move.y = y;
-	fz_transform_point(&arg->move, arg->ctm);
+	arg->move = fz_transform_point_xy(x, y, arg->ctm);
 	arg->trailing_move = 1;
 }
 
@@ -852,11 +889,7 @@ static void
 bound_lineto(fz_context *ctx, void *arg_, float x, float y)
 {
 	bound_path_arg *arg = (bound_path_arg *)arg_;
-	fz_point p;
-
-	p.x = x;
-	p.y = y;
-	fz_transform_point(&p, arg->ctm);
+	fz_point p = fz_transform_point_xy(x, y, arg->ctm);
 	if (arg->first)
 	{
 		arg->rect.x0 = arg->rect.x1 = p.x;
@@ -864,12 +897,11 @@ bound_lineto(fz_context *ctx, void *arg_, float x, float y)
 		arg->first = 0;
 	}
 	else
-		bound_expand(&arg->rect, &p);
-
+		bound_expand(&arg->rect, p);
 	if (arg->trailing_move)
 	{
 		arg->trailing_move = 0;
-		bound_expand(&arg->rect, &arg->move);
+		bound_expand(&arg->rect, arg->move);
 	}
 }
 
@@ -877,11 +909,7 @@ static void
 bound_curveto(fz_context *ctx, void *arg_, float x1, float y1, float x2, float y2, float x3, float y3)
 {
 	bound_path_arg *arg = (bound_path_arg *)arg_;
-	fz_point p;
-
-	p.x = x1;
-	p.y = y1;
-	fz_transform_point(&p, arg->ctm);
+	fz_point p = fz_transform_point_xy(x1, y1, arg->ctm);
 	if (arg->first)
 	{
 		arg->rect.x0 = arg->rect.x1 = p.x;
@@ -889,17 +917,13 @@ bound_curveto(fz_context *ctx, void *arg_, float x1, float y1, float x2, float y
 		arg->first = 0;
 	}
 	else
-		bound_expand(&arg->rect, &p);
-	p.x = x2;
-	p.y = y2;
-	bound_expand(&arg->rect, fz_transform_point(&p, arg->ctm));
-	p.x = x3;
-	p.y = y3;
-	bound_expand(&arg->rect, fz_transform_point(&p, arg->ctm));
+		bound_expand(&arg->rect, p);
+	bound_expand(&arg->rect, fz_transform_point_xy(x2, y2, arg->ctm));
+	bound_expand(&arg->rect, fz_transform_point_xy(x3, y3, arg->ctm));
 	if (arg->trailing_move)
 	{
 		arg->trailing_move = 0;
-		bound_expand(&arg->rect, &arg->move);
+		bound_expand(&arg->rect, arg->move);
 	}
 }
 
@@ -911,8 +935,8 @@ static const fz_path_walker bound_path_walker =
 	NULL
 };
 
-fz_rect *
-fz_bound_path(fz_context *ctx, const fz_path *path, const fz_stroke_state *stroke, const fz_matrix *ctm, fz_rect *r)
+fz_rect
+fz_bound_path(fz_context *ctx, const fz_path *path, const fz_stroke_state *stroke, fz_matrix ctm)
 {
 	bound_path_arg arg;
 
@@ -925,15 +949,14 @@ fz_bound_path(fz_context *ctx, const fz_path *path, const fz_stroke_state *strok
 
 	if (!arg.first && stroke)
 	{
-		fz_adjust_rect_for_stroke(ctx, &arg.rect, stroke, ctm);
+		arg.rect = fz_adjust_rect_for_stroke(ctx, arg.rect, stroke, ctm);
 	}
 
-	*r = arg.rect;
-	return r;
+	return arg.rect;
 }
 
-fz_rect *
-fz_adjust_rect_for_stroke(fz_context *ctx, fz_rect *r, const fz_stroke_state *stroke, const fz_matrix *ctm)
+fz_rect
+fz_adjust_rect_for_stroke(fz_context *ctx, fz_rect r, const fz_stroke_state *stroke, fz_matrix ctm)
 {
 	float expand;
 
@@ -947,15 +970,15 @@ fz_adjust_rect_for_stroke(fz_context *ctx, fz_rect *r, const fz_stroke_state *st
 	if ((stroke->linejoin == FZ_LINEJOIN_MITER || stroke->linejoin == FZ_LINEJOIN_MITER_XPS) && stroke->miterlimit > 1)
 		expand *= stroke->miterlimit;
 
-	r->x0 -= expand;
-	r->y0 -= expand;
-	r->x1 += expand;
-	r->y1 += expand;
+	r.x0 -= expand;
+	r.y0 -= expand;
+	r.x1 += expand;
+	r.y1 += expand;
 	return r;
 }
 
 void
-fz_transform_path(fz_context *ctx, fz_path *path, const fz_matrix *ctm)
+fz_transform_path(fz_context *ctx, fz_path *path, fz_matrix ctm)
 {
 	int i, k, n;
 	fz_point p, p1, p2, p3, q, s;
@@ -963,7 +986,7 @@ fz_transform_path(fz_context *ctx, fz_path *path, const fz_matrix *ctm)
 	if (path->packed)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "Cannot transform a packed path");
 
-	if (ctm->b == 0 && ctm->c == 0)
+	if (ctm.b == 0 && ctm.c == 0)
 	{
 		/* Simple, in place transform */
 		i = 0;
@@ -1004,16 +1027,14 @@ fz_transform_path(fz_context *ctx, fz_path *path, const fz_matrix *ctm)
 			case FZ_HORIZTO:
 			case FZ_HORIZTOCLOSE:
 				q.x = path->coords[k];
-				p = q;
-				fz_transform_point(&p, ctm);
+				p = fz_transform_point(q, ctm);
 				path->coords[k++] = p.x;
 				n = 0;
 				break;
 			case FZ_VERTTO:
 			case FZ_VERTTOCLOSE:
 				q.y = path->coords[k];
-				p = q;
-				fz_transform_point(&p, ctm);
+				p = fz_transform_point(q, ctm);
 				path->coords[k++] = p.y;
 				n = 0;
 				break;
@@ -1024,8 +1045,7 @@ fz_transform_path(fz_context *ctx, fz_path *path, const fz_matrix *ctm)
 			{
 				q.x = path->coords[k];
 				q.y = path->coords[k+1];
-				p = q;
-				fz_transform_point(&p, ctm);
+				p = fz_transform_point(q, ctm);
 				path->coords[k++] = p.x;
 				path->coords[k++] = p.y;
 				n--;
@@ -1051,7 +1071,7 @@ fz_transform_path(fz_context *ctx, fz_path *path, const fz_matrix *ctm)
 			i++;
 		}
 	}
-	else if (ctm->a == 0 && ctm->d == 0)
+	else if (ctm.a == 0 && ctm.d == 0)
 	{
 		/* In place transform with command rewriting */
 		i = 0;
@@ -1091,32 +1111,28 @@ fz_transform_path(fz_context *ctx, fz_path *path, const fz_matrix *ctm)
 				break;
 			case FZ_HORIZTO:
 				q.x = path->coords[k];
-				p = q;
-				fz_transform_point(&p, ctm);
+				p = fz_transform_point(q, ctm);
 				path->coords[k++] = p.y;
 				path->cmds[i] = FZ_VERTTO;
 				n = 0;
 				break;
 			case FZ_HORIZTOCLOSE:
 				q.x = path->coords[k];
-				p = q;
-				fz_transform_point(&p, ctm);
+				p = fz_transform_point(q, ctm);
 				path->coords[k++] = p.y;
 				path->cmds[i] = FZ_VERTTOCLOSE;
 				n = 0;
 				break;
 			case FZ_VERTTO:
 				q.y = path->coords[k];
-				p = q;
-				fz_transform_point(&p, ctm);
+				p = fz_transform_point(q, ctm);
 				path->coords[k++] = p.x;
 				path->cmds[i] = FZ_HORIZTO;
 				n = 0;
 				break;
 			case FZ_VERTTOCLOSE:
 				q.y = path->coords[k];
-				p = q;
-				fz_transform_point(&p, ctm);
+				p = fz_transform_point(q, ctm);
 				path->coords[k++] = p.x;
 				path->cmds[i] = FZ_HORIZTOCLOSE;
 				n = 0;
@@ -1126,10 +1142,9 @@ fz_transform_path(fz_context *ctx, fz_path *path, const fz_matrix *ctm)
 			}
 			while (n > 0)
 			{
-				p.x = path->coords[k];
-				p.y = path->coords[k+1];
-				q = p;
-				fz_transform_point(&p, ctm);
+				q.x = path->coords[k];
+				q.y = path->coords[k+1];
+				p = fz_transform_point(q, ctm);
 				path->coords[k++] = p.x;
 				path->coords[k++] = p.y;
 				n--;
@@ -1185,12 +1200,12 @@ fz_transform_path(fz_context *ctx, fz_path *path, const fz_matrix *ctm)
 		}
 		if (path->cmd_len + extra_cmd < path->cmd_cap)
 		{
-			path->cmds = fz_resize_array(ctx, path->cmds, path->cmd_len + extra_cmd, sizeof(unsigned char));
+			path->cmds = fz_realloc_array(ctx, path->cmds, path->cmd_len + extra_cmd, unsigned char);
 			path->cmd_cap = path->cmd_len + extra_cmd;
 		}
 		if (path->coord_len + extra_coord < path->coord_cap)
 		{
-			path->coords = fz_resize_array(ctx, path->coords, path->coord_len + extra_coord, sizeof(float));
+			path->coords = fz_realloc_array(ctx, path->coords, path->coord_len + extra_coord, float);
 			path->coord_cap = path->coord_len + extra_coord;
 		}
 		memmove(path->cmds + extra_cmd, path->cmds, path->cmd_len * sizeof(unsigned char));
@@ -1236,10 +1251,10 @@ fz_transform_path(fz_context *ctx, fz_path *path, const fz_matrix *ctm)
 				p3.x = p.x;
 				p3.y = p2.y;
 				s = p;
-				fz_transform_point(&p, ctm);
-				fz_transform_point(&p1, ctm);
-				fz_transform_point(&p2, ctm);
-				fz_transform_point(&p3, ctm);
+				p = fz_transform_point(p, ctm);
+				p1 = fz_transform_point(p1, ctm);
+				p2 = fz_transform_point(p2, ctm);
+				p3 = fz_transform_point(p3, ctm);
 				path->coords[coord_write++] = p.x;
 				path->coords[coord_write++] = p.y;
 				path->coords[coord_write++] = p1.x;
@@ -1256,8 +1271,7 @@ fz_transform_path(fz_context *ctx, fz_path *path, const fz_matrix *ctm)
 				break;
 			case FZ_HORIZTO:
 				q.x = path->coords[coord_read++];
-				p = q;
-				fz_transform_point(&p, ctm);
+				p = fz_transform_point(q, ctm);
 				path->coords[coord_write++] = p.x;
 				path->coords[coord_write++] = p.y;
 				path->cmds[cmd_write-1] = FZ_LINETO;
@@ -1266,7 +1280,7 @@ fz_transform_path(fz_context *ctx, fz_path *path, const fz_matrix *ctm)
 			case FZ_HORIZTOCLOSE:
 				p.x = path->coords[coord_read++];
 				p.y = q.y;
-				fz_transform_point(&p, ctm);
+				p = fz_transform_point(p, ctm);
 				path->coords[coord_write++] = p.x;
 				path->coords[coord_write++] = p.y;
 				path->cmds[cmd_write-1] = FZ_LINETOCLOSE;
@@ -1275,8 +1289,7 @@ fz_transform_path(fz_context *ctx, fz_path *path, const fz_matrix *ctm)
 				break;
 			case FZ_VERTTO:
 				q.y = path->coords[coord_read++];
-				p = q;
-				fz_transform_point(&p, ctm);
+				p = fz_transform_point(q, ctm);
 				path->coords[coord_write++] = p.x;
 				path->coords[coord_write++] = p.y;
 				path->cmds[cmd_write-1] = FZ_LINETO;
@@ -1285,7 +1298,7 @@ fz_transform_path(fz_context *ctx, fz_path *path, const fz_matrix *ctm)
 			case FZ_VERTTOCLOSE:
 				p.x = q.x;
 				p.y = path->coords[coord_read++];
-				fz_transform_point(&p, ctm);
+				p = fz_transform_point(p, ctm);
 				path->coords[coord_write++] = p.x;
 				path->coords[coord_write++] = p.y;
 				path->cmds[cmd_write-1] = FZ_LINETOCLOSE;
@@ -1297,10 +1310,9 @@ fz_transform_path(fz_context *ctx, fz_path *path, const fz_matrix *ctm)
 			}
 			while (n > 0)
 			{
-				p.x = path->coords[coord_read++];
-				p.y = path->coords[coord_read++];
-				q = p;
-				fz_transform_point(&p, ctm);
+				q.x = path->coords[coord_read++];
+				q.y = path->coords[coord_read++];
+				p = fz_transform_point(q, ctm);
 				path->coords[coord_write++] = p.x;
 				path->coords[coord_write++] = p.y;
 				n--;
@@ -1333,91 +1345,16 @@ void fz_trim_path(fz_context *ctx, fz_path *path)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "Can't trim a packed path");
 	if (path->cmd_cap > path->cmd_len)
 	{
-		path->cmds = fz_resize_array(ctx, path->cmds, path->cmd_len, sizeof(unsigned char));
+		path->cmds = fz_realloc_array(ctx, path->cmds, path->cmd_len, unsigned char);
 		path->cmd_cap = path->cmd_len;
 	}
 	if (path->coord_cap > path->coord_len)
 	{
-		path->coords = fz_resize_array(ctx, path->coords, path->coord_len, sizeof(float));
+		path->coords = fz_realloc_array(ctx, path->coords, path->coord_len, float);
 		path->coord_cap = path->coord_len;
 	}
 }
 
-void
-fz_print_path(fz_context *ctx, fz_output *out, fz_path *path, int indent)
-{
-	float x, y;
-	int i = 0, k = 0;
-	int n;
-	while (i < path->cmd_len)
-	{
-		uint8_t cmd = path->cmds[i++];
-
-		for (n = 0; n < indent; n++)
-			fz_write_byte(ctx, out, ' ');
-		switch (cmd)
-		{
-		case FZ_MOVETO:
-		case FZ_MOVETOCLOSE:
-			x = path->coords[k++];
-			y = path->coords[k++];
-			fz_write_printf(ctx, out, "%g %g m%s\n", x, y, cmd == FZ_MOVETOCLOSE ? " z" : "");
-			break;
-		case FZ_LINETO:
-		case FZ_LINETOCLOSE:
-			x = path->coords[k++];
-			y = path->coords[k++];
-			fz_write_printf(ctx, out, "%g %g l%s\n", x, y, cmd == FZ_LINETOCLOSE ? " z" : "");
-			break;
-		case FZ_DEGENLINETO:
-		case FZ_DEGENLINETOCLOSE:
-			fz_write_printf(ctx, out, "d%s\n", cmd == FZ_DEGENLINETOCLOSE ? " z" : "");
-			break;
-		case FZ_HORIZTO:
-		case FZ_HORIZTOCLOSE:
-			x = path->coords[k++];
-			fz_write_printf(ctx, out, "%g h%s\n", x, cmd == FZ_HORIZTOCLOSE ? " z" : "");
-			break;
-		case FZ_VERTTOCLOSE:
-		case FZ_VERTTO:
-			y = path->coords[k++];
-			fz_write_printf(ctx, out, "%g i%s\n", y, cmd == FZ_VERTTOCLOSE ? " z" : "");
-			break;
-		case FZ_CURVETOCLOSE:
-		case FZ_CURVETO:
-			x = path->coords[k++];
-			y = path->coords[k++];
-			fz_write_printf(ctx, out, "%g %g ", x, y);
-			x = path->coords[k++];
-			y = path->coords[k++];
-			fz_write_printf(ctx, out, "%g %g ", x, y);
-			x = path->coords[k++];
-			y = path->coords[k++];
-			fz_write_printf(ctx, out, "%g %g c%s\n", x, y, cmd == FZ_CURVETOCLOSE ? " z" : "");
-			break;
-		case FZ_CURVETOVCLOSE:
-		case FZ_CURVETOV:
-		case FZ_CURVETOYCLOSE:
-		case FZ_CURVETOY:
-			x = path->coords[k++];
-			y = path->coords[k++];
-			fz_write_printf(ctx, out, "%g %g ", x, y);
-			x = path->coords[k++];
-			y = path->coords[k++];
-			fz_write_printf(ctx, out, "%g %g %c%s\n", x, y, (cmd == FZ_CURVETOVCLOSE || cmd == FZ_CURVETOV ? 'v' : 'y'), (cmd == FZ_CURVETOVCLOSE || cmd == FZ_CURVETOYCLOSE) ? " z" : "");
-			break;
-		case FZ_RECTTO:
-			x = path->coords[k++];
-			y = path->coords[k++];
-			fz_write_printf(ctx, out, "%g %g ", x, y);
-			x = path->coords[k++];
-			y = path->coords[k++];
-			fz_write_printf(ctx, out, "%g %g r\n", x, y);
-			break;
-		}
-	}
-}
-
 const fz_stroke_state fz_default_stroke_state = {
 	-2, /* -2 is the magic number we use when we have stroke states stored on the stack */
 	FZ_LINECAP_BUTT, FZ_LINECAP_BUTT, FZ_LINECAP_BUTT,
@@ -1561,10 +1498,10 @@ fz_clone_path(fz_context *ctx, fz_path *path)
 		case FZ_PATH_PACKED_OPEN:
 			new_path->cmd_len = path->cmd_len;
 			new_path->cmd_cap = path->cmd_cap;
-			new_path->cmds = clone_block(ctx, path->cmds, path->cmd_cap);
+			new_path->cmds = Memento_label(clone_block(ctx, path->cmds, path->cmd_cap), "path_cmds");
 			new_path->coord_len = path->coord_len;
 			new_path->coord_cap = path->coord_cap;
-			new_path->coords = clone_block(ctx, path->coords, sizeof(float)*path->coord_cap);
+			new_path->coords = Memento_label(clone_block(ctx, path->coords, sizeof(float)*path->coord_cap), "path_coords");
 			new_path->current = path->current;
 			new_path->begin = path->begin;
 			break;
@@ -1580,9 +1517,9 @@ fz_clone_path(fz_context *ctx, fz_path *path)
 				new_path->coord_len = ppath->coord_len;
 				new_path->coord_cap = ppath->coord_len;
 				data = (uint8_t *)&ppath[1];
-				new_path->coords = clone_block(ctx, data, sizeof(float)*path->coord_cap);
+				new_path->coords = Memento_label(clone_block(ctx, data, sizeof(float)*path->coord_cap), "path_coords");
 				data += sizeof(float) * path->coord_cap;
-				new_path->cmds = clone_block(ctx, data, path->cmd_cap);
+				new_path->cmds = Memento_label(clone_block(ctx, data, path->cmd_cap), "path_cmds");
 				xy = new_path->coords;
 				for (i = 0; i < new_path->cmd_len; i++)
 				{
@@ -1644,6 +1581,7 @@ fz_clone_path(fz_context *ctx, fz_path *path)
 	}
 	fz_catch(ctx)
 	{
+		fz_free(ctx, new_path->coords);
 		fz_free(ctx, new_path->cmds);
 		fz_free(ctx, new_path);
 		fz_rethrow(ctx);
