@@ -1,5 +1,26 @@
 #include "mupdf/fitz.h"
 
+#include <float.h>
+#include <math.h>
+#include <stdarg.h>
+#include <stdio.h>
+
+#ifdef _MSC_VER
+#if _MSC_VER < 1500 /* MSVC 2008 */
+int snprintf(char *s, size_t n, const char *fmt, ...)
+{
+		int r;
+		va_list ap;
+		va_start(ap, fmt);
+		r = vsprintf(s, fmt, ap);
+		va_end(ap);
+		return r;
+}
+#else if _MSC_VER < 1900 /* MSVC 2015 */
+#define snprintf _snprintf
+#endif
+#endif
+
 static const char *fz_hex_digits = "0123456789abcdef";
 
 struct fmtbuf
@@ -88,10 +109,14 @@ static void fmtuint32(struct fmtbuf *out, unsigned int a, int s, int z, int w, i
 		buf[i++] = fz_hex_digits[a % base];
 		a /= base;
 	}
+	if (s) {
+		if (z == '0')
+			while (i < w - 1)
+				buf[i++] = z;
+		buf[i++] = s;
+	}
 	while (i < w)
 		buf[i++] = z;
-	if (s)
-		fmtputc(out, '+');
 	while (i > 0)
 		fmtputc(out, buf[--i]);
 }
@@ -108,10 +133,14 @@ static void fmtuint64(struct fmtbuf *out, uint64_t a, int s, int z, int w, int b
 		buf[i++] = fz_hex_digits[a % base];
 		a /= base;
 	}
+	if (s) {
+		if (z == '0')
+			while (i < w - 1)
+				buf[i++] = z;
+		buf[i++] = s;
+	}
 	while (i < w)
 		buf[i++] = z;
-	if (s)
-		fmtputc(out, '+');
 	while (i > 0)
 		fmtputc(out, buf[--i]);
 }
@@ -122,11 +151,19 @@ static void fmtint32(struct fmtbuf *out, int value, int s, int z, int w, int bas
 
 	if (value < 0)
 	{
-		fmtputc(out, '-');
+		s = '-';
 		a = -value;
 	}
-	else
+	else if (s)
+	{
+		s = '+';
 		a = value;
+	}
+	else
+	{
+		s = 0;
+		a = value;
+	}
 	fmtuint32(out, a, s, z, w, base);
 }
 
@@ -136,26 +173,89 @@ static void fmtint64(struct fmtbuf *out, int64_t value, int s, int z, int w, int
 
 	if (value < 0)
 	{
-		fmtputc(out, '-');
+		s = '-';
 		a = -value;
 	}
-	else
+	else if (s)
+	{
+		s = '+';
 		a = value;
+	}
+	else
+	{
+		s = 0;
+		a = value;
+	}
 	fmtuint64(out, a, s, z, w, base);
 }
 
-static void fmtquote(struct fmtbuf *out, const char *s, int sq, int eq)
+static void fmtquote(struct fmtbuf *out, const char *s, int sq, int eq, int verbatim)
+{
+	int i, n, c;
+	fmtputc(out, sq);
+	while (*s != 0) {
+		n = fz_chartorune(&c, s);
+		switch (c) {
+		default:
+			if (c < 32) {
+				fmtputc(out, '\\');
+				fmtputc(out, 'x');
+				fmtputc(out, "0123456789ABCDEF"[(c>>4)&15]);
+				fmtputc(out, "0123456789ABCDEF"[(c)&15]);
+			} else if (c > 127) {
+				if (verbatim)
+				{
+					for (i = 0; i < n; ++i)
+						fmtputc(out, s[i]);
+				}
+				else
+				{
+					fmtputc(out, '\\');
+					fmtputc(out, 'u');
+					fmtputc(out, "0123456789ABCDEF"[(c>>12)&15]);
+					fmtputc(out, "0123456789ABCDEF"[(c>>8)&15]);
+					fmtputc(out, "0123456789ABCDEF"[(c>>4)&15]);
+					fmtputc(out, "0123456789ABCDEF"[(c)&15]);
+				}
+			} else {
+				if (c == sq || c == eq)
+					fmtputc(out, '\\');
+				fmtputc(out, c);
+			}
+			break;
+		case '\\': fmtputc(out, '\\'); fmtputc(out, '\\'); break;
+		case '\b': fmtputc(out, '\\'); fmtputc(out, 'b'); break;
+		case '\f': fmtputc(out, '\\'); fmtputc(out, 'f'); break;
+		case '\n': fmtputc(out, '\\'); fmtputc(out, 'n'); break;
+		case '\r': fmtputc(out, '\\'); fmtputc(out, 'r'); break;
+		case '\t': fmtputc(out, '\\'); fmtputc(out, 't'); break;
+		}
+		s += n;
+	}
+	fmtputc(out, eq);
+}
+
+static void fmtquote_pdf(struct fmtbuf *out, const char *s, int sq, int eq)
 {
 	int c;
 	fmtputc(out, sq);
-	while ((c = *s++) != 0) {
+	while ((c = (unsigned char)*s++) != 0) {
 		switch (c) {
 		default:
 			if (c < 32 || c > 127) {
 				fmtputc(out, '\\');
-				fmtputc(out, '0' + ((c >> 6) & 7));
-				fmtputc(out, '0' + ((c >> 3) & 7));
-				fmtputc(out, '0' + ((c) & 7));
+				if (sq == '(')
+				{
+					fmtputc(out, '0' + ((c >> 6) & 7));
+					fmtputc(out, '0' + ((c >> 3) & 7));
+					fmtputc(out, '0' + ((c) & 7));
+				}
+				else
+				{
+					fmtputc(out, 'x');
+					fmtputc(out, "0123456789ABCDEF"[(c>>4)&15]);
+					fmtputc(out, "0123456789ABCDEF"[(c)&15]);
+				}
 			} else {
 				if (c == sq || c == eq)
 					fmtputc(out, '\\');
@@ -173,6 +273,21 @@ static void fmtquote(struct fmtbuf *out, const char *s, int sq, int eq)
 	fmtputc(out, eq);
 }
 
+static void fmtname(struct fmtbuf *out, const char *s)
+{
+	int c;
+	fmtputc(out, '/');
+	while ((c = *s++) != 0) {
+		if (c <= 32 || c == '/' || c == '#') {
+			fmtputc(out, '#');
+			fmtputc(out, "0123456789ABCDEF"[(c>>4)&15]);
+			fmtputc(out, "0123456789ABCDEF"[(c)&15]);
+		} else {
+			fmtputc(out, c);
+		}
+	}
+}
+
 void
 fz_format_string(fz_context *ctx, void *user, void (*emit)(fz_context *ctx, void *user, int c), const char *fmt, va_list args)
 {
@@ -191,30 +306,28 @@ fz_format_string(fz_context *ctx, void *user, void (*emit)(fz_context *ctx, void
 	{
 		if (c == '%')
 		{
-			c = *fmt++;
+			s = 0;
+			z = ' ';
+
+			/* flags */
+			while ((c = *fmt++) != 0)
+			{
+				/* plus sign */
+				if (c == '+')
+					s = 1;
+				/* space sign */
+				else if (c == ' ')
+					s = ' ';
+				/* zero padding */
+				else if (c == '0')
+					z = '0';
+				/* TODO: '-' to left justify */
+				else
+					break;
+			}
 			if (c == 0)
 				break;
 
-			/* sign */
-			s = 0;
-			if (c == '+') {
-				s = 1;
-				c = *fmt++;
-				if (c == 0)
-					break;
-			}
-
-			/* TODO: '-' to left justify */
-
-			/* leading zero */
-			z = ' ';
-			if (c == '0') {
-				z = '0';
-				c = *fmt++;
-				if (c == 0)
-					break;
-			}
-
 			/* width */
 			w = 0;
 			if (c == '*') {
@@ -254,11 +367,7 @@ fz_format_string(fz_context *ctx, void *user, void (*emit)(fz_context *ctx, void
 			bits = 0;
 			if (c == 'l') {
 				c = *fmt++;
-				bits = sizeof(long) * 8;
-				if (c == 'l') {
-					c = *fmt++;
-					bits = 64;
-				}
+				bits = sizeof(int64_t) * 8;
 				if (c == 0)
 					break;
 			}
@@ -274,12 +383,6 @@ fz_format_string(fz_context *ctx, void *user, void (*emit)(fz_context *ctx, void
 				if (c == 0)
 					break;
 			}
-			if (c == 'Z') {
-				c = *fmt++;
-				bits = sizeof(fz_off_t) * 8;
-				if (c == 0)
-					break;
-			}
 
 			switch (c) {
 			default:
@@ -346,7 +449,7 @@ fz_format_string(fz_context *ctx, void *user, void (*emit)(fz_context *ctx, void
 
 			case 'p':
 				bits = 8 * sizeof(void *);
-				w = 2 * sizeof(void *);
+				z = '0';
 				fmtputc(&out, '0');
 				fmtputc(&out, 'x');
 				/* fallthrough */
@@ -354,15 +457,16 @@ fz_format_string(fz_context *ctx, void *user, void (*emit)(fz_context *ctx, void
 				if (bits == 64)
 				{
 					i64 = va_arg(args, int64_t);
-					fmtuint64(&out, i64, s, z, w, 16);
+					fmtuint64(&out, i64, 0, z, w, 16);
 				}
 				else
 				{
 					i32 = va_arg(args, int);
-					fmtuint32(&out, i32, s, z, w, 16);
+					fmtuint32(&out, i32, 0, z, w, 16);
 				}
 				break;
 			case 'd':
+			case 'i':
 				if (bits == 64)
 				{
 					i64 = va_arg(args, int64_t);
@@ -378,18 +482,14 @@ fz_format_string(fz_context *ctx, void *user, void (*emit)(fz_context *ctx, void
 				if (bits == 64)
 				{
 					i64 = va_arg(args, int64_t);
-					fmtuint64(&out, i64, s, z, w, 10);
+					fmtuint64(&out, i64, 0, z, w, 10);
 				}
 				else
 				{
 					i32 = va_arg(args, int);
-					fmtuint32(&out, i32, s, z, w, 10);
+					fmtuint32(&out, i32, 0, z, w, 10);
 				}
 				break;
-			case 'o':
-				i32 = va_arg(args, int);
-				fmtint32(&out, i32, s, z, w, 8);
-				break;
 
 			case 's':
 				str = va_arg(args, const char*);
@@ -398,15 +498,25 @@ fz_format_string(fz_context *ctx, void *user, void (*emit)(fz_context *ctx, void
 				while ((c = *str++) != 0)
 					fmtputc(&out, c);
 				break;
+			case 'Q': /* quoted string (with verbatim unicode) */
+				str = va_arg(args, const char*);
+				if (!str) str = "";
+				fmtquote(&out, str, '"', '"', 1);
+				break;
 			case 'q': /* quoted string */
 				str = va_arg(args, const char*);
 				if (!str) str = "";
-				fmtquote(&out, str, '"', '"');
+				fmtquote(&out, str, '"', '"', 0);
 				break;
 			case '(': /* pdf string */
 				str = va_arg(args, const char*);
 				if (!str) str = "";
-				fmtquote(&out, str, '(', ')');
+				fmtquote_pdf(&out, str, '(', ')');
+				break;
+			case 'n': /* pdf name */
+				str = va_arg(args, const char*);
+				if (!str) str = "";
+				fmtname(&out, str);
 				break;
 			}
 		}
@@ -436,13 +546,15 @@ fz_vsnprintf(char *buffer, size_t space, const char *fmt, va_list args)
 {
 	struct snprintf_buffer out;
 	out.p = buffer;
-	out.s = space;
+	out.s = space > 0 ? space - 1 : 0;
 	out.n = 0;
 
 	/* Note: using a NULL context is safe here */
 	fz_format_string(NULL, &out, snprintf_emit, fmt, args);
-	snprintf_emit(NULL, &out, 0);
-	return out.n - 1;
+	if (space > 0)
+		out.p[out.n < space ? out.n : space - 1] = '\0';
+
+	return out.n;
 }
 
 size_t
@@ -451,14 +563,31 @@ fz_snprintf(char *buffer, size_t space, const char *fmt, ...)
 	va_list ap;
 	struct snprintf_buffer out;
 	out.p = buffer;
-	out.s = space;
+	out.s = space > 0 ? space - 1 : 0;
 	out.n = 0;
 
 	va_start(ap, fmt);
 	/* Note: using a NULL context is safe here */
 	fz_format_string(NULL, &out, snprintf_emit, fmt, ap);
-	snprintf_emit(NULL, &out, 0);
+	if (space > 0)
+		out.p[out.n < space ? out.n : space - 1] = '\0';
 	va_end(ap);
 
-	return out.n - 1;
+	return out.n;
+}
+
+char *
+fz_asprintf(fz_context *ctx, const char *fmt, ...)
+{
+	size_t len;
+	char *mem;
+	va_list ap;
+	va_start(ap, fmt);
+	len = fz_vsnprintf(NULL, 0, fmt, ap);
+	va_end(ap);
+	mem = Memento_label(fz_malloc(ctx, len+1), "asprintf");
+	va_start(ap, fmt);
+	fz_vsnprintf(mem, len+1, fmt, ap);
+	va_end(ap);
+	return mem;
 }
