@@ -3,15 +3,14 @@
 #include <string.h>
 #include <math.h>
 
-typedef struct fz_mesh_processor_s fz_mesh_processor;
-
-struct fz_mesh_processor_s {
+typedef struct
+{
 	fz_shade *shade;
 	fz_shade_prepare_fn *prepare;
 	fz_shade_process_fn *process;
 	void *process_arg;
 	int ncomp;
-};
+} fz_mesh_processor;
 
 #define SWAP(a,b) {fz_vertex *t = (a); (a) = (b); (b) = t;}
 
@@ -442,8 +441,8 @@ fz_process_shade_type5(fz_context *ctx, fz_shade *shade, fz_matrix ctm, fz_mesh_
 
 	fz_try(ctx)
 	{
-		ref = fz_malloc_array(ctx, vprow, sizeof(fz_vertex));
-		buf = fz_malloc_array(ctx, vprow, sizeof(fz_vertex));
+		ref = fz_malloc_array(ctx, vprow, fz_vertex);
+		buf = fz_malloc_array(ctx, vprow, fz_vertex);
 		first = 1;
 
 		while (!fz_is_eof_bits(ctx, stream))
@@ -479,28 +478,26 @@ fz_process_shade_type5(fz_context *ctx, fz_shade *shade, fz_matrix ctm, fz_mesh_
 
 /* Subdivide and tessellate tensor-patches */
 
-typedef struct tensor_patch_s tensor_patch;
-
-struct tensor_patch_s
+typedef struct
 {
 	fz_point pole[4][4];
 	float color[4][FZ_MAX_COLORS];
-};
+} tensor_patch;
 
 static void
-triangulate_patch(fz_context *ctx, fz_mesh_processor *painter, tensor_patch p)
+triangulate_patch(fz_context *ctx, fz_mesh_processor *painter, tensor_patch *p)
 {
 	fz_vertex v0, v1, v2, v3;
 
-	v0.p = p.pole[0][0];
-	v1.p = p.pole[0][3];
-	v2.p = p.pole[3][3];
-	v3.p = p.pole[3][0];
+	v0.p = p->pole[0][0];
+	v1.p = p->pole[0][3];
+	v2.p = p->pole[3][3];
+	v3.p = p->pole[3][0];
 
-	fz_prepare_color(ctx, painter, &v0, p.color[0]);
-	fz_prepare_color(ctx, painter, &v1, p.color[1]);
-	fz_prepare_color(ctx, painter, &v2, p.color[2]);
-	fz_prepare_color(ctx, painter, &v3, p.color[3]);
+	fz_prepare_color(ctx, painter, &v0, p->color[0]);
+	fz_prepare_color(ctx, painter, &v1, p->color[1]);
+	fz_prepare_color(ctx, painter, &v2, p->color[2]);
+	fz_prepare_color(ctx, painter, &v3, p->color[3]);
 
 	paint_quad(ctx, painter, &v0, &v1, &v2, &v3);
 }
@@ -584,8 +581,8 @@ draw_stripe(fz_context *ctx, fz_mesh_processor *painter, tensor_patch *p, int de
 	if (depth == 0)
 	{
 		/* if no more subdividing, draw two new patches... */
-		triangulate_patch(ctx, painter, s1);
-		triangulate_patch(ctx, painter, s0);
+		triangulate_patch(ctx, painter, &s1);
+		triangulate_patch(ctx, painter, &s0);
 	}
 	else
 	{
@@ -792,6 +789,7 @@ fz_process_shade_type6(fz_context *ctx, fz_shade *shade, fz_matrix ctm, fz_mesh_
 
 			if (flag == 0)
 			{
+				/* No patch data to copy forwards */
 			}
 			else if (flag == 1 && prevc)
 			{
@@ -905,6 +903,7 @@ fz_process_shade_type7(fz_context *ctx, fz_shade *shade, fz_matrix ctm, fz_mesh_
 
 			if (flag == 0)
 			{
+				/* No patch data to copy forward */
 			}
 			else if (flag == 1 && prevc)
 			{
@@ -958,27 +957,6 @@ fz_process_shade_type7(fz_context *ctx, fz_shade *shade, fz_matrix ctm, fz_mesh_
 	}
 }
 
-/*
-	Process a shade, using supplied callback
-	functions. This decomposes the shading to a mesh (even ones
-	that are not natively meshes, such as linear or radial
-	shadings), and processes triangles from those meshes.
-
-	shade: The shade to process.
-
-	ctm: The transform to use
-
-	prepare: Callback function to 'prepare' each vertex.
-	This function is passed an array of floats, and populates
-	a fz_vertex structure.
-
-	process: This function is passed 3 pointers to vertex
-	structures, and actually performs the processing (typically
-	filling the area between the vertexes).
-
-	process_arg: An opaque argument passed through from caller
-	to callback functions.
-*/
 void
 fz_process_shade(fz_context *ctx, fz_shade *shade, fz_matrix ctm, fz_rect scissor,
 		fz_shade_prepare_fn *prepare, fz_shade_process_fn *process, void *process_arg)
@@ -1103,12 +1081,6 @@ fz_keep_shade(fz_context *ctx, fz_shade *shade)
 	return fz_keep_storable(ctx, &shade->storable);
 }
 
-/*
-	Internal function to destroy a
-	shade. Only exposed for use with the fz_store.
-
-	shade: The reference to destroy.
-*/
 void
 fz_drop_shade_imp(fz_context *ctx, fz_storable *shade_)
 {
@@ -1127,17 +1099,6 @@ fz_drop_shade(fz_context *ctx, fz_shade *shade)
 	fz_drop_storable(ctx, &shade->storable);
 }
 
-/*
-	Bound a given shading.
-
-	shade: The shade to bound.
-
-	ctm: The transform to apply to the shade before bounding.
-
-	r: Pointer to storage to put the bounds in.
-
-	Returns r, updated to contain the bounds for the shading.
-*/
 fz_rect
 fz_bound_shade(fz_context *ctx, fz_shade *shade, fz_matrix ctm)
 {
