@@ -5,16 +5,65 @@
 #include <float.h>
 #include <string.h>
 
+/* Simple layout structure */
+
+fz_layout_block *fz_new_layout(fz_context *ctx)
+{
+	fz_pool *pool = fz_new_pool(ctx);
+	fz_layout_block *block;
+	fz_try(ctx)
+	{
+		block = fz_pool_alloc(ctx, pool, sizeof (fz_layout_block));
+		block->pool = pool;
+		block->head = NULL;
+		block->tailp = &block->head;
+	}
+	fz_catch(ctx)
+	{
+		fz_drop_pool(ctx, pool);
+		fz_rethrow(ctx);
+	}
+	return block;
+}
+
+void fz_drop_layout(fz_context *ctx, fz_layout_block *block)
+{
+	if (block)
+		fz_drop_pool(ctx, block->pool);
+}
+
+void fz_add_layout_line(fz_context *ctx, fz_layout_block *block, float x, float y, float h, const char *p)
+{
+	fz_layout_line *line = fz_pool_alloc(ctx, block->pool, sizeof (fz_layout_line));
+	line->x = x;
+	line->y = y;
+	line->h = h;
+	line->p = p;
+	line->text = NULL;
+	line->next = NULL;
+	*block->tailp = line;
+	block->tailp = &line->next;
+	block->text_tailp = &line->text;
+}
+
+void fz_add_layout_char(fz_context *ctx, fz_layout_block *block, float x, float w, const char *p)
+{
+	fz_layout_char *ch = fz_pool_alloc(ctx, block->pool, sizeof (fz_layout_char));
+	ch->x = x;
+	ch->w = w;
+	ch->p = p;
+	ch->next = NULL;
+	*block->text_tailp = ch;
+	block->text_tailp = &ch->next;
+}
+
 /* Extract text into blocks and lines. */
 
-#define LINE_DIST 0.9f
+#define PARAGRAPH_DIST 1.5f
 #define SPACE_DIST 0.15f
 #define SPACE_MAX_DIST 0.8f
-#define PARAGRAPH_DIST 0.5f
 
-typedef struct fz_stext_device_s fz_stext_device;
-
-struct fz_stext_device_s
+typedef struct
 {
 	fz_device super;
 	fz_stext_page *page;
@@ -24,17 +73,22 @@ struct fz_stext_device_s
 	int curdir;
 	int lastchar;
 	int flags;
-};
+	int color;
+	const fz_text *lasttext;
+} fz_stext_device;
 
 const char *fz_stext_options_usage =
 	"Text output options:\n"
+	"\tinhibit-spaces: don't add spaces between gaps in the text\n"
+	"\tpreserve-images: keep images in output\n"
 	"\tpreserve-ligatures: do not expand ligatures into constituent characters\n"
 	"\tpreserve-whitespace: do not convert all whitespace into space characters\n"
-	"\tpreserve-images: keep images in output\n"
+	"\tpreserve-spans: do not merge spans on the same line\n"
+	"\tdehyphenate: attempt to join up hyphenated words\n"
 	"\n";
 
 fz_stext_page *
-fz_new_stext_page(fz_context *ctx, const fz_rect *mediabox)
+fz_new_stext_page(fz_context *ctx, fz_rect mediabox)
 {
 	fz_pool *pool = fz_new_pool(ctx);
 	fz_stext_page *page = NULL;
@@ -42,7 +96,7 @@ fz_new_stext_page(fz_context *ctx, const fz_rect *mediabox)
 	{
 		page = fz_pool_alloc(ctx, pool, sizeof(*page));
 		page->pool = pool;
-		page->mediabox = *mediabox;
+		page->mediabox = mediabox;
 		page->first_block = NULL;
 		page->last_block = NULL;
 	}
@@ -91,17 +145,13 @@ add_text_block_to_page(fz_context *ctx, fz_stext_page *page)
 }
 
 static fz_stext_block *
-add_image_block_to_page(fz_context *ctx, fz_stext_page *page, const fz_matrix *ctm, fz_image *image)
+add_image_block_to_page(fz_context *ctx, fz_stext_page *page, fz_matrix ctm, fz_image *image)
 {
 	fz_stext_block *block = add_block_to_page(ctx, page);
 	block->type = FZ_STEXT_BLOCK_IMAGE;
-	block->u.i.transform = *ctm;
+	block->u.i.transform = ctm;
 	block->u.i.image = fz_keep_image(ctx, image);
-	block->bbox.x0 = 0;
-	block->bbox.y0 = 0;
-	block->bbox.x1 = 1;
-	block->bbox.y1 = 1;
-	fz_transform_rect(&block->bbox, ctm);
+	block->bbox = fz_transform_rect(fz_unit_rect, ctm);
 	return block;
 }
 
@@ -124,18 +174,8 @@ add_line_to_block(fz_context *ctx, fz_stext_page *page, fz_stext_block *block, c
 	return line;
 }
 
-static float min4(float a, float b, float c, float d)
-{
-	return fz_min(fz_min(a, b), fz_min(c, d));
-}
-
-static float max4(float a, float b, float c, float d)
-{
-	return fz_max(fz_max(a, b), fz_max(c, d));
-}
-
 static fz_stext_char *
-add_char_to_line(fz_context *ctx, fz_stext_page *page, fz_stext_line *line, const fz_matrix *trm, fz_font *font, float size, int c, fz_point *p, fz_point *q)
+add_char_to_line(fz_context *ctx, fz_stext_page *page, fz_stext_line *line, fz_matrix trm, fz_font *font, float size, int c, fz_point *p, fz_point *q, int color)
 {
 	fz_stext_char *ch = fz_pool_alloc(ctx, page->pool, sizeof *line->first_char);
 	fz_point a, d;
@@ -149,6 +189,7 @@ add_char_to_line(fz_context *ctx, fz_stext_page *page, fz_stext_line *line, cons
 	}
 
 	ch->c = c;
+	ch->color = color;
 	ch->origin = *p;
 	ch->size = size;
 	ch->font = font; /* TODO: keep and drop */
@@ -162,23 +203,44 @@ add_char_to_line(fz_context *ctx, fz_stext_page *page, fz_stext_line *line, cons
 	}
 	else
 	{
-		fz_rect *bbox = fz_font_bbox(ctx, font);
-		a.x = bbox->x1;
-		d.x = bbox->x0;
+		fz_rect bbox = fz_font_bbox(ctx, font);
+		a.x = bbox.x1;
+		d.x = bbox.x0;
 		a.y = 0;
 		d.y = 0;
 	}
-	fz_transform_vector(&a, trm);
-	fz_transform_vector(&d, trm);
+	a = fz_transform_vector(a, trm);
+	d = fz_transform_vector(d, trm);
 
-	ch->bbox.x0 = min4(p->x + a.x, q->x + a.x, p->x + d.x, q->x + d.x);
-	ch->bbox.x1 = max4(p->x + a.x, q->x + a.x, p->x + d.x, q->x + d.x);
-	ch->bbox.y0 = min4(p->y + a.y, q->y + a.y, p->y + d.y, q->y + d.y);
-	ch->bbox.y1 = max4(p->y + a.y, q->y + a.y, p->y + d.y, q->y + d.y);
+	ch->quad.ll = fz_make_point(p->x + d.x, p->y + d.y);
+	ch->quad.ul = fz_make_point(p->x + a.x, p->y + a.y);
+	ch->quad.lr = fz_make_point(q->x + d.x, q->y + d.y);
+	ch->quad.ur = fz_make_point(q->x + a.x, q->y + a.y);
 
 	return ch;
 }
 
+static void
+remove_last_char(fz_context *ctx, fz_stext_line *line)
+{
+	if (line && line->first_char)
+	{
+		fz_stext_char *prev = NULL;
+		fz_stext_char *ch = line->first_char;
+		while (ch->next)
+		{
+			prev = ch;
+			ch = ch->next;
+		}
+		if (prev)
+		{
+			/* the characters are pool allocated, so we don't actually leak the removed node */
+			line->last_char = prev;
+			line->last_char->next = NULL;
+		}
+	}
+}
+
 static int
 direction_from_bidi_class(int bidiclass, int curdir)
 {
@@ -212,6 +274,12 @@ direction_from_bidi_class(int bidiclass, int curdir)
 	}
 }
 
+static int is_hyphen(int c)
+{
+	/* check for: hyphen-minus, soft hyphen, hyphen, and non-breaking hyphen */
+	return (c == '-' || c == 0xAD || c == 0x2010 || c == 0x2011);
+}
+
 static float
 vec_dot(const fz_point *a, const fz_point *b)
 {
@@ -219,7 +287,7 @@ vec_dot(const fz_point *a, const fz_point *b)
 }
 
 static void
-fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int c, int glyph, fz_matrix *trm, float adv, int wmode)
+fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int c, int glyph, fz_matrix trm, float adv, int wmode, int force_new_line)
 {
 	fz_stext_page *page = dev->page;
 	fz_stext_block *cur_block;
@@ -248,9 +316,8 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 		dir.x = 0;
 		dir.y = -1;
 	}
-	fz_transform_vector(&dir, trm);
-	ndir = dir;
-	fz_normalize_vector(&ndir);
+	dir = fz_transform_vector(dir, trm);
+	ndir = fz_normalize_vector(dir);
 
 	size = fz_matrix_expansion(trm);
 
@@ -271,17 +338,17 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 	 */
 	if (wmode == 0)
 	{
-		p.x = trm->e;
-		p.y = trm->f;
-		q.x = trm->e + adv * dir.x;
-		q.y = trm->f + adv * dir.y;
+		p.x = trm.e;
+		p.y = trm.f;
+		q.x = trm.e + adv * dir.x;
+		q.y = trm.f + adv * dir.y;
 	}
 	else
 	{
-		p.x = trm->e - adv * dir.x;
-		p.y = trm->f - adv * dir.y;
-		q.x = trm->e;
-		q.y = trm->f;
+		p.x = trm.e - adv * dir.x;
+		p.y = trm.f - adv * dir.y;
+		q.x = trm.e;
+		q.y = trm.f;
 	}
 
 	/* Find current position to enter new text. */
@@ -293,7 +360,7 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 	if (cur_line && glyph < 0)
 	{
 		/* Don't advance pen or break lines for no-glyph characters in a cluster */
-		add_char_to_line(ctx, page, cur_line, trm, font, size, c, &dev->pen, &dev->pen);
+		add_char_to_line(ctx, page, cur_line, trm, font, size, c, &dev->pen, &dev->pen, dev->color);
 		dev->lastchar = c;
 		return;
 	}
@@ -332,20 +399,25 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 			{
 				if (fabsf(spacing) < size * SPACE_DIST)
 				{
-					/* Motion is in line, and small. */
+					/* Motion is in line and small enough to ignore. */
 					new_line = 0;
 				}
-				else if (spacing >= size * SPACE_DIST && spacing < size * SPACE_MAX_DIST)
+				else if (fabsf(spacing) > size * SPACE_MAX_DIST)
 				{
-					/* Motion is in line, but large enough to warrant us adding a space. */
-					if (dev->lastchar != ' ' && wmode == 0)
-						add_space = 1;
+					/* Motion is in line and large enough to warrant splitting to a new line */
+					new_line = 1;
+				}
+				else if (spacing < 0)
+				{
+					/* Motion is backward in line! Ignore this odd spacing. */
 					new_line = 0;
 				}
 				else
 				{
-					/* Motion is in line, but large enough to warrant splitting to a new line */
-					new_line = 1;
+					/* Motion is forward in line and large enough to warrant us adding a space. */
+					if (dev->lastchar != ' ' && wmode == 0)
+						add_space = 1;
+					new_line = 0;
 				}
 			}
 
@@ -361,7 +433,7 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 		}
 
 		/* Enough for a new line, but not enough for a new paragraph */
-		else if (fabsf(base_offset) < size * 1.3f)
+		else if (fabsf(base_offset) <= size * PARAGRAPH_DIST)
 		{
 			/* Check indent to spot text-indent style paragraphs */
 			if (wmode == 0 && cur_line && dev->new_obj)
@@ -385,27 +457,33 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 		cur_line = cur_block->u.t.last_line;
 	}
 
+	if (new_line && (dev->flags & FZ_STEXT_DEHYPHENATE) && is_hyphen(dev->lastchar))
+	{
+		remove_last_char(ctx, cur_line);
+		new_line = 0;
+	}
+
 	/* Start a new line */
-	if (new_line || !cur_line)
+	if (new_line || !cur_line || force_new_line)
 	{
 		cur_line = add_line_to_block(ctx, page, cur_block, &ndir, wmode);
 		dev->start = p;
 	}
 
 	/* Add synthetic space */
-	if (add_space)
-		add_char_to_line(ctx, page, cur_line, trm, font, size, ' ', &dev->pen, &p);
+	if (add_space && !(dev->flags & FZ_STEXT_INHIBIT_SPACES))
+		add_char_to_line(ctx, page, cur_line, trm, font, size, ' ', &dev->pen, &p, dev->color);
 
-	add_char_to_line(ctx, page, cur_line, trm, font, size, c, &p, &q);
+	add_char_to_line(ctx, page, cur_line, trm, font, size, c, &p, &q, dev->color);
 	dev->lastchar = c;
 	dev->pen = q;
 
 	dev->new_obj = 0;
-	dev->trm = *trm;
+	dev->trm = trm;
 }
 
 static void
-fz_add_stext_char(fz_context *ctx, fz_stext_device *dev, fz_font *font, int c, int glyph, fz_matrix *trm, float adv, int wmode)
+fz_add_stext_char(fz_context *ctx, fz_stext_device *dev, fz_font *font, int c, int glyph, fz_matrix trm, float adv, int wmode, int force_new_line)
 {
 	/* ignore when one unicode character maps to multiple glyphs */
 	if (c == -1)
@@ -416,31 +494,31 @@ fz_add_stext_char(fz_context *ctx, fz_stext_device *dev, fz_font *font, int c, i
 		switch (c)
 		{
 		case 0xFB00: /* ff */
-			fz_add_stext_char_imp(ctx, dev, font, 'f', glyph, trm, adv, wmode);
-			fz_add_stext_char_imp(ctx, dev, font, 'f', -1, trm, 0, wmode);
+			fz_add_stext_char_imp(ctx, dev, font, 'f', glyph, trm, adv, wmode, force_new_line);
+			fz_add_stext_char_imp(ctx, dev, font, 'f', -1, trm, 0, wmode, 0);
 			return;
 		case 0xFB01: /* fi */
-			fz_add_stext_char_imp(ctx, dev, font, 'f', glyph, trm, adv, wmode);
-			fz_add_stext_char_imp(ctx, dev, font, 'i', -1, trm, 0, wmode);
+			fz_add_stext_char_imp(ctx, dev, font, 'f', glyph, trm, adv, wmode, force_new_line);
+			fz_add_stext_char_imp(ctx, dev, font, 'i', -1, trm, 0, wmode, 0);
 			return;
 		case 0xFB02: /* fl */
-			fz_add_stext_char_imp(ctx, dev, font, 'f', glyph, trm, adv, wmode);
-			fz_add_stext_char_imp(ctx, dev, font, 'l', -1, trm, 0, wmode);
+			fz_add_stext_char_imp(ctx, dev, font, 'f', glyph, trm, adv, wmode, force_new_line);
+			fz_add_stext_char_imp(ctx, dev, font, 'l', -1, trm, 0, wmode, 0);
 			return;
 		case 0xFB03: /* ffi */
-			fz_add_stext_char_imp(ctx, dev, font, 'f', glyph, trm, adv, wmode);
-			fz_add_stext_char_imp(ctx, dev, font, 'f', -1, trm, 0, wmode);
-			fz_add_stext_char_imp(ctx, dev, font, 'i', -1, trm, 0, wmode);
+			fz_add_stext_char_imp(ctx, dev, font, 'f', glyph, trm, adv, wmode, force_new_line);
+			fz_add_stext_char_imp(ctx, dev, font, 'f', -1, trm, 0, wmode, 0);
+			fz_add_stext_char_imp(ctx, dev, font, 'i', -1, trm, 0, wmode, 0);
 			return;
 		case 0xFB04: /* ffl */
-			fz_add_stext_char_imp(ctx, dev, font, 'f', glyph, trm, adv, wmode);
-			fz_add_stext_char_imp(ctx, dev, font, 'f', -1, trm, 0, wmode);
-			fz_add_stext_char_imp(ctx, dev, font, 'l', -1, trm, 0, wmode);
+			fz_add_stext_char_imp(ctx, dev, font, 'f', glyph, trm, adv, wmode, force_new_line);
+			fz_add_stext_char_imp(ctx, dev, font, 'f', -1, trm, 0, wmode, 0);
+			fz_add_stext_char_imp(ctx, dev, font, 'l', -1, trm, 0, wmode, 0);
 			return;
 		case 0xFB05: /* long st */
 		case 0xFB06: /* st */
-			fz_add_stext_char_imp(ctx, dev, font, 's', glyph, trm, adv, wmode);
-			fz_add_stext_char_imp(ctx, dev, font, 't', -1, trm, 0, wmode);
+			fz_add_stext_char_imp(ctx, dev, font, 's', glyph, trm, adv, wmode, force_new_line);
+			fz_add_stext_char_imp(ctx, dev, font, 't', -1, trm, 0, wmode, 0);
 			return;
 		}
 	}
@@ -472,11 +550,11 @@ fz_add_stext_char(fz_context *ctx, fz_stext_device *dev, fz_font *font, int c, i
 		}
 	}
 
-	fz_add_stext_char_imp(ctx, dev, font, c, glyph, trm, adv, wmode);
+	fz_add_stext_char_imp(ctx, dev, font, c, glyph, trm, adv, wmode, force_new_line);
 }
 
 static void
-fz_stext_extract(fz_context *ctx, fz_stext_device *dev, fz_text_span *span, const fz_matrix *ctm)
+fz_stext_extract(fz_context *ctx, fz_stext_device *dev, fz_text_span *span, fz_matrix ctm)
 {
 	fz_font *font = span->font;
 	fz_matrix tm = span->trm;
@@ -487,16 +565,12 @@ fz_stext_extract(fz_context *ctx, fz_stext_device *dev, fz_text_span *span, cons
 	if (span->len == 0)
 		return;
 
-	tm.e = 0;
-	tm.f = 0;
-	fz_concat(&trm, &tm, ctm);
-
 	for (i = 0; i < span->len; i++)
 	{
 		/* Calculate new pen location and delta */
 		tm.e = span->items[i].x;
 		tm.f = span->items[i].y;
-		fz_concat(&trm, &tm, ctm);
+		trm = fz_concat(tm, ctm);
 
 		/* Calculate bounding box and new pen position based on font metrics */
 		if (span->items[i].gid >= 0)
@@ -504,66 +578,107 @@ fz_stext_extract(fz_context *ctx, fz_stext_device *dev, fz_text_span *span, cons
 		else
 			adv = 0;
 
-		fz_add_stext_char(ctx, dev, font, span->items[i].ucs, span->items[i].gid, &trm, adv, span->wmode);
+		fz_add_stext_char(ctx, dev, font,
+			span->items[i].ucs,
+			span->items[i].gid,
+			trm,
+			adv,
+			span->wmode,
+			(i == 0) && (dev->flags & FZ_STEXT_PRESERVE_SPANS));
 	}
 }
 
-static void
-fz_stext_fill_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_matrix *ctm,
-	fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
+static int hexrgb_from_color(fz_context *ctx, fz_colorspace *colorspace, const float *color)
 {
-	fz_stext_device *tdev = (fz_stext_device*)dev;
-	fz_text_span *span;
-	tdev->new_obj = 1;
-	for (span = text->head; span; span = span->next)
-		fz_stext_extract(ctx, tdev, span, ctm);
+	float rgb[3];
+	fz_convert_color(ctx, colorspace, color, fz_device_rgb(ctx), rgb, NULL, fz_default_color_params);
+	return
+		(fz_clampi(rgb[0] * 255, 0, 255) << 16) |
+		(fz_clampi(rgb[1] * 255, 0, 255) << 8) |
+		(fz_clampi(rgb[2] * 255, 0, 255));
 }
 
 static void
-fz_stext_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_stroke_state *stroke, const fz_matrix *ctm,
-	fz_colorspace *colorspace, const float *color, float alpha, const fz_color_params *color_params)
+fz_stext_fill_text(fz_context *ctx, fz_device *dev, const fz_text *text, fz_matrix ctm,
+	fz_colorspace *colorspace, const float *color, float alpha, fz_color_params color_params)
 {
 	fz_stext_device *tdev = (fz_stext_device*)dev;
 	fz_text_span *span;
+	if (text == tdev->lasttext)
+		return;
+	tdev->color = hexrgb_from_color(ctx, colorspace, color);
 	tdev->new_obj = 1;
 	for (span = text->head; span; span = span->next)
 		fz_stext_extract(ctx, tdev, span, ctm);
+	fz_drop_text(ctx, tdev->lasttext);
+	tdev->lasttext = fz_keep_text(ctx, text);
 }
 
 static void
-fz_stext_clip_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_matrix *ctm, const fz_rect *scissor)
+fz_stext_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_stroke_state *stroke, fz_matrix ctm,
+	fz_colorspace *colorspace, const float *color, float alpha, fz_color_params color_params)
 {
 	fz_stext_device *tdev = (fz_stext_device*)dev;
 	fz_text_span *span;
+	if (text == tdev->lasttext)
+		return;
+	tdev->color = hexrgb_from_color(ctx, colorspace, color);
 	tdev->new_obj = 1;
 	for (span = text->head; span; span = span->next)
 		fz_stext_extract(ctx, tdev, span, ctm);
+	fz_drop_text(ctx, tdev->lasttext);
+	tdev->lasttext = fz_keep_text(ctx, text);
 }
 
 static void
-fz_stext_clip_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_stroke_state *stroke, const fz_matrix *ctm, const fz_rect *scissor)
+fz_stext_clip_text(fz_context *ctx, fz_device *dev, const fz_text *text, fz_matrix ctm, fz_rect scissor)
 {
 	fz_stext_device *tdev = (fz_stext_device*)dev;
 	fz_text_span *span;
+	if (text == tdev->lasttext)
+		return;
+	tdev->color = 0;
 	tdev->new_obj = 1;
 	for (span = text->head; span; span = span->next)
 		fz_stext_extract(ctx, tdev, span, ctm);
+	fz_drop_text(ctx, tdev->lasttext);
+	tdev->lasttext = fz_keep_text(ctx, text);
 }
 
 static void
-fz_stext_ignore_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_matrix *ctm)
+fz_stext_clip_stroke_text(fz_context *ctx, fz_device *dev, const fz_text *text, const fz_stroke_state *stroke, fz_matrix ctm, fz_rect scissor)
 {
 	fz_stext_device *tdev = (fz_stext_device*)dev;
 	fz_text_span *span;
+	if (text == tdev->lasttext)
+		return;
+	tdev->color = 0;
 	tdev->new_obj = 1;
 	for (span = text->head; span; span = span->next)
 		fz_stext_extract(ctx, tdev, span, ctm);
+	fz_drop_text(ctx, tdev->lasttext);
+	tdev->lasttext = fz_keep_text(ctx, text);
+}
+
+static void
+fz_stext_ignore_text(fz_context *ctx, fz_device *dev, const fz_text *text, fz_matrix ctm)
+{
+	fz_stext_device *tdev = (fz_stext_device*)dev;
+	fz_text_span *span;
+	if (text == tdev->lasttext)
+		return;
+	tdev->color = 0;
+	tdev->new_obj = 1;
+	for (span = text->head; span; span = span->next)
+		fz_stext_extract(ctx, tdev, span, ctm);
+	fz_drop_text(ctx, tdev->lasttext);
+	tdev->lasttext = fz_keep_text(ctx, text);
 }
 
 /* Images and shadings */
 
 static void
-fz_stext_fill_image(fz_context *ctx, fz_device *dev, fz_image *img, const fz_matrix *ctm, float alpha, const fz_color_params *color_params)
+fz_stext_fill_image(fz_context *ctx, fz_device *dev, fz_image *img, fz_matrix ctm, float alpha, fz_color_params color_params)
 {
 	fz_stext_device *tdev = (fz_stext_device*)dev;
 
@@ -575,14 +690,14 @@ fz_stext_fill_image(fz_context *ctx, fz_device *dev, fz_image *img, const fz_mat
 }
 
 static void
-fz_stext_fill_image_mask(fz_context *ctx, fz_device *dev, fz_image *img, const fz_matrix *ctm,
-		fz_colorspace *cspace, const float *color, float alpha, const fz_color_params *color_params)
+fz_stext_fill_image_mask(fz_context *ctx, fz_device *dev, fz_image *img, fz_matrix ctm,
+		fz_colorspace *cspace, const float *color, float alpha, fz_color_params color_params)
 {
 	fz_stext_fill_image(ctx, dev, img, ctm, alpha, color_params);
 }
 
 static fz_image *
-fz_new_image_from_shade(fz_context *ctx, fz_shade *shade, fz_matrix *in_out_ctm, const fz_color_params *color_params, const fz_rect *scissor)
+fz_new_image_from_shade(fz_context *ctx, fz_shade *shade, fz_matrix *in_out_ctm, fz_color_params color_params, fz_rect scissor)
 {
 	fz_matrix ctm = *in_out_ctm;
 	fz_pixmap *pix;
@@ -590,18 +705,18 @@ fz_new_image_from_shade(fz_context *ctx, fz_shade *shade, fz_matrix *in_out_ctm,
 	fz_rect bounds;
 	fz_irect bbox;
 
-	fz_bound_shade(ctx, shade, &ctm, &bounds);
-	fz_intersect_rect(&bounds, scissor);
-	fz_irect_from_rect(&bbox, &bounds);
+	bounds = fz_bound_shade(ctx, shade, ctm);
+	bounds = fz_intersect_rect(bounds, scissor);
+	bbox = fz_irect_from_rect(bounds);
 
-	pix = fz_new_pixmap_with_bbox(ctx, fz_device_rgb(ctx), &bbox, NULL, !shade->use_background);
+	pix = fz_new_pixmap_with_bbox(ctx, fz_device_rgb(ctx), bbox, NULL, !shade->use_background);
 	fz_try(ctx)
 	{
 		if (shade->use_background)
 			fz_fill_pixmap_with_color(ctx, pix, shade->colorspace, shade->background, color_params);
 		else
 			fz_clear_pixmap(ctx, pix);
-		fz_paint_shade(ctx, shade, NULL, &ctm, pix, color_params, &bbox, NULL);
+		fz_paint_shade(ctx, shade, NULL, ctm, pix, color_params, bbox, NULL);
 		img = fz_new_image_from_pixmap(ctx, pix, NULL);
 	}
 	fz_always(ctx)
@@ -619,13 +734,13 @@ fz_new_image_from_shade(fz_context *ctx, fz_shade *shade, fz_matrix *in_out_ctm,
 }
 
 static void
-fz_stext_fill_shade(fz_context *ctx, fz_device *dev, fz_shade *shade, const fz_matrix *ctm, float alpha, const fz_color_params *color_params)
+fz_stext_fill_shade(fz_context *ctx, fz_device *dev, fz_shade *shade, fz_matrix ctm, float alpha, fz_color_params color_params)
 {
-	fz_matrix local_ctm = *ctm;
-	const fz_rect *scissor = fz_device_current_scissor(ctx, dev);
+	fz_matrix local_ctm = ctm;
+	fz_rect scissor = fz_device_current_scissor(ctx, dev);
 	fz_image *image = fz_new_image_from_shade(ctx, shade, &local_ctm, color_params, scissor);
 	fz_try(ctx)
-		fz_stext_fill_image(ctx, dev, image, &local_ctm, alpha, color_params);
+		fz_stext_fill_image(ctx, dev, image, local_ctm, alpha, color_params);
 	fz_always(ctx)
 		fz_drop_image(ctx, image);
 	fz_catch(ctx)
@@ -648,8 +763,14 @@ fz_stext_close_device(fz_context *ctx, fz_device *dev)
 		for (line = block->u.t.first_line; line; line = line->next)
 		{
 			for (ch = line->first_char; ch; ch = ch->next)
-				fz_union_rect(&line->bbox, &ch->bbox);
-			fz_union_rect(&block->bbox, &line->bbox);
+			{
+				fz_rect ch_box = fz_rect_from_quad(ch->quad);
+				if (ch == line->first_char)
+					line->bbox = ch_box;
+				else
+					line->bbox = fz_union_rect(line->bbox, ch_box);
+			}
+			block->bbox = fz_union_rect(block->bbox, line->bbox);
 		}
 	}
 
@@ -660,6 +781,8 @@ fz_stext_close_device(fz_context *ctx, fz_device *dev)
 static void
 fz_stext_drop_device(fz_context *ctx, fz_device *dev)
 {
+	fz_stext_device *tdev = (fz_stext_device*)dev;
+	fz_drop_text(ctx, tdev->lasttext);
 }
 
 fz_stext_options *
@@ -675,6 +798,12 @@ fz_parse_stext_options(fz_context *ctx, fz_stext_options *opts, const char *stri
 		opts->flags |= FZ_STEXT_PRESERVE_WHITESPACE;
 	if (fz_has_option(ctx, string, "preserve-images", &val) && fz_option_eq(val, "yes"))
 		opts->flags |= FZ_STEXT_PRESERVE_IMAGES;
+	if (fz_has_option(ctx, string, "inhibit-spaces", &val) && fz_option_eq(val, "yes"))
+		opts->flags |= FZ_STEXT_INHIBIT_SPACES;
+	if (fz_has_option(ctx, string, "dehyphenate", &val) && fz_option_eq(val, "yes"))
+		opts->flags |= FZ_STEXT_DEHYPHENATE;
+	if (fz_has_option(ctx, string, "preserve-spans", &val) && fz_option_eq(val, "yes"))
+		opts->flags |= FZ_STEXT_PRESERVE_SPANS;
 
 	return opts;
 }
@@ -695,18 +824,20 @@ fz_new_stext_device(fz_context *ctx, fz_stext_page *page, const fz_stext_options
 
 	if (opts && (opts->flags & FZ_STEXT_PRESERVE_IMAGES))
 	{
-		dev->super.hints |= FZ_MAINTAIN_CONTAINER_STACK;
 		dev->super.fill_shade = fz_stext_fill_shade;
 		dev->super.fill_image = fz_stext_fill_image;
 		dev->super.fill_image_mask = fz_stext_fill_image_mask;
 	}
 
+	if (opts)
+		dev->flags = opts->flags;
 	dev->page = page;
 	dev->pen.x = 0;
 	dev->pen.y = 0;
 	dev->trm = fz_identity;
 	dev->lastchar = ' ';
 	dev->curdir = 1;
+	dev->lasttext = NULL;
 
 	return (fz_device*)dev;
 }
