@@ -1,6 +1,5 @@
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
-#include "pdf-imp.h"
 
 #include <string.h>
 #include <math.h>
@@ -64,25 +63,16 @@ pdf_clear_stack(fz_context *ctx, pdf_csi *csi)
 }
 
 static pdf_font_desc *
-load_font_or_hail_mary(fz_context *ctx, pdf_document *doc, pdf_obj *rdb, pdf_obj *font, int depth, fz_cookie *cookie)
+pdf_try_load_font(fz_context *ctx, pdf_document *doc, pdf_obj *rdb, pdf_obj *font, fz_cookie *cookie)
 {
-	pdf_font_desc *desc;
-
+	pdf_font_desc *desc = NULL;
 	fz_try(ctx)
-	{
-		desc = pdf_load_font(ctx, doc, rdb, font, depth);
-	}
+		desc = pdf_load_font(ctx, doc, rdb, font);
 	fz_catch(ctx)
 	{
-		if (fz_caught(ctx) == FZ_ERROR_TRYLATER && cookie && cookie->incomplete_ok)
-		{
-			desc = NULL;
-			cookie->incomplete++;
-		}
-		else
-		{
-			fz_rethrow(ctx);
-		}
+		if (fz_caught(ctx) == FZ_ERROR_TRYLATER)
+			if (cookie)
+				cookie->incomplete++;
 	}
 	if (desc == NULL)
 		desc = pdf_load_hail_mary_font(ctx, doc);
@@ -90,11 +80,12 @@ load_font_or_hail_mary(fz_context *ctx, pdf_document *doc, pdf_obj *rdb, pdf_obj
 }
 
 static fz_image *
-parse_inline_image(fz_context *ctx, pdf_csi *csi, fz_stream *stm)
+parse_inline_image(fz_context *ctx, pdf_csi *csi, fz_stream *stm, char *csname, int cslen)
 {
 	pdf_document *doc = csi->doc;
 	pdf_obj *rdb = csi->rdb;
 	pdf_obj *obj = NULL;
+	pdf_obj *cs;
 	fz_image *img = NULL;
 	int ch, found;
 
@@ -105,6 +96,15 @@ parse_inline_image(fz_context *ctx, pdf_csi *csi, fz_stream *stm)
 	{
 		obj = pdf_parse_dict(ctx, doc, stm, &doc->lexbuf.base);
 
+		if (csname)
+		{
+			cs = pdf_dict_get(ctx, obj, PDF_NAME(CS));
+			if (!pdf_is_indirect(ctx, cs) && pdf_is_name(ctx, cs))
+				fz_strlcpy(csname, pdf_to_name(ctx, cs), cslen);
+			else
+				csname[0] = 0;
+		}
+
 		/* read whitespace after ID keyword */
 		ch = fz_read_byte(ctx, stm);
 		if (ch == '\r')
@@ -126,7 +126,7 @@ parse_inline_image(fz_context *ctx, pdf_csi *csi, fz_stream *stm)
 				if (ch == 'I')
 				{
 					ch = fz_peek_byte(ctx, stm);
-					if (ch == ' ' || ch <= 32 || ch == EOF || ch == '<' || ch == '/')
+					if (ch == ' ' || ch <= 32 || ch == '<' || ch == '/')
 					{
 						found = 1;
 						break;
@@ -155,23 +155,23 @@ pdf_process_extgstate(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, pdf_ob
 {
 	pdf_obj *obj;
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_LW);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(LW));
 	if (pdf_is_number(ctx, obj) && proc->op_w)
 		proc->op_w(ctx, proc, pdf_to_real(ctx, obj));
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_LC);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(LC));
 	if (pdf_is_int(ctx, obj) && proc->op_J)
 		proc->op_J(ctx, proc, fz_clampi(pdf_to_int(ctx, obj), 0, 2));
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_LJ);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(LJ));
 	if (pdf_is_int(ctx, obj) && proc->op_j)
 		proc->op_j(ctx, proc, fz_clampi(pdf_to_int(ctx, obj), 0, 2));
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_ML);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(ML));
 	if (pdf_is_number(ctx, obj) && proc->op_M)
 		proc->op_M(ctx, proc, pdf_to_real(ctx, obj));
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_D);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(D));
 	if (pdf_is_array(ctx, obj) && proc->op_d)
 	{
 		pdf_obj *dash_array = pdf_array_get(ctx, obj, 0);
@@ -179,20 +179,24 @@ pdf_process_extgstate(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, pdf_ob
 		proc->op_d(ctx, proc, dash_array, pdf_to_real(ctx, dash_phase));
 	}
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_RI);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(RI));
 	if (pdf_is_name(ctx, obj) && proc->op_ri)
 		proc->op_ri(ctx, proc, pdf_to_name(ctx, obj));
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_FL);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(FL));
 	if (pdf_is_number(ctx, obj) && proc->op_i)
 		proc->op_i(ctx, proc, pdf_to_real(ctx, obj));
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_Font);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(Font));
 	if (pdf_is_array(ctx, obj) && proc->op_Tf)
 	{
 		pdf_obj *font_ref = pdf_array_get(ctx, obj, 0);
 		pdf_obj *font_size = pdf_array_get(ctx, obj, 1);
-		pdf_font_desc *font = load_font_or_hail_mary(ctx, csi->doc, csi->rdb, font_ref, 0, csi->cookie);
+		pdf_font_desc *font;
+		if (pdf_is_dict(ctx, font_ref))
+			font = pdf_try_load_font(ctx, csi->doc, csi->rdb, font_ref, csi->cookie);
+		else
+			font = pdf_load_hail_mary_font(ctx, csi->doc);
 		fz_try(ctx)
 			proc->op_Tf(ctx, proc, "ExtGState", font, pdf_to_real(ctx, font_size));
 		fz_always(ctx)
@@ -203,53 +207,53 @@ pdf_process_extgstate(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, pdf_ob
 
 	/* overprint and color management */
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_OP);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(OP));
 	if (pdf_is_bool(ctx, obj) && proc->op_gs_OP)
 		proc->op_gs_OP(ctx, proc, pdf_to_bool(ctx, obj));
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_op);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(op));
 	if (pdf_is_bool(ctx, obj) && proc->op_gs_op)
 		proc->op_gs_op(ctx, proc, pdf_to_bool(ctx, obj));
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_OPM);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(OPM));
 	if (pdf_is_int(ctx, obj) && proc->op_gs_OPM)
 		proc->op_gs_OPM(ctx, proc, pdf_to_int(ctx, obj));
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_UseBlackPtComp);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(UseBlackPtComp));
 	if (pdf_is_name(ctx, obj) && proc->op_gs_UseBlackPtComp)
 		proc->op_gs_UseBlackPtComp(ctx, proc, obj);
 
 	/* transfer functions */
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_TR2);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(TR2));
 	if (pdf_is_name(ctx, obj))
-		if (!pdf_name_eq(ctx, obj, PDF_NAME_Identity) && !pdf_name_eq(ctx, obj, PDF_NAME_Default))
+		if (!pdf_name_eq(ctx, obj, PDF_NAME(Identity)) && !pdf_name_eq(ctx, obj, PDF_NAME(Default)))
 			fz_warn(ctx, "ignoring transfer function");
 	if (!obj) /* TR is ignored in the presence of TR2 */
 	{
-		pdf_obj *tr = pdf_dict_get(ctx, dict, PDF_NAME_TR);
+		pdf_obj *tr = pdf_dict_get(ctx, dict, PDF_NAME(TR));
 		if (pdf_is_name(ctx, tr))
-			if (!pdf_name_eq(ctx, tr, PDF_NAME_Identity))
+			if (!pdf_name_eq(ctx, tr, PDF_NAME(Identity)))
 				fz_warn(ctx, "ignoring transfer function");
 	}
 
 	/* transparency state */
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_CA);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(CA));
 	if (pdf_is_number(ctx, obj) && proc->op_gs_CA)
 		proc->op_gs_CA(ctx, proc, pdf_to_real(ctx, obj));
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_ca);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(ca));
 	if (pdf_is_number(ctx, obj) && proc->op_gs_ca)
 		proc->op_gs_ca(ctx, proc, pdf_to_real(ctx, obj));
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_BM);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(BM));
 	if (pdf_is_array(ctx, obj))
 		obj = pdf_array_get(ctx, obj, 0);
 	if (pdf_is_name(ctx, obj) && proc->op_gs_BM)
 		proc->op_gs_BM(ctx, proc, pdf_to_name(ctx, obj));
 
-	obj = pdf_dict_get(ctx, dict, PDF_NAME_SMask);
+	obj = pdf_dict_get(ctx, dict, PDF_NAME(SMask));
 	if (proc->op_gs_SMask)
 	{
 		if (pdf_is_dict(ctx, obj))
@@ -260,7 +264,7 @@ pdf_process_extgstate(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, pdf_ob
 			int colorspace_n = 1;
 			int k, luminosity;
 
-			xobj = pdf_dict_get(ctx, obj, PDF_NAME_G);
+			xobj = pdf_dict_get(ctx, obj, PDF_NAME(G));
 
 			colorspace = pdf_xobject_colorspace(ctx, xobj);
 			if (colorspace)
@@ -276,26 +280,26 @@ pdf_process_extgstate(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, pdf_ob
 				softmask_bc[3] = 1.0f;
 			fz_drop_colorspace(ctx, colorspace);
 
-			bc = pdf_dict_get(ctx, obj, PDF_NAME_BC);
+			bc = pdf_dict_get(ctx, obj, PDF_NAME(BC));
 			if (pdf_is_array(ctx, bc))
 			{
 				for (k = 0; k < colorspace_n; k++)
-					softmask_bc[k] = pdf_to_real(ctx, pdf_array_get(ctx, bc, k));
+					softmask_bc[k] = pdf_array_get_real(ctx, bc, k);
 			}
 
-			s = pdf_dict_get(ctx, obj, PDF_NAME_S);
-			if (pdf_name_eq(ctx, s, PDF_NAME_Luminosity))
+			s = pdf_dict_get(ctx, obj, PDF_NAME(S));
+			if (pdf_name_eq(ctx, s, PDF_NAME(Luminosity)))
 				luminosity = 1;
 			else
 				luminosity = 0;
 
-			tr = pdf_dict_get(ctx, obj, PDF_NAME_TR);
-			if (tr && !pdf_name_eq(ctx, tr, PDF_NAME_Identity))
+			tr = pdf_dict_get(ctx, obj, PDF_NAME(TR));
+			if (tr && !pdf_name_eq(ctx, tr, PDF_NAME(Identity)))
 				fz_warn(ctx, "ignoring transfer function");
 
 			proc->op_gs_SMask(ctx, proc, xobj, csi->rdb, softmask_bc, luminosity);
 		}
-		else if (pdf_is_name(ctx, obj) && pdf_name_eq(ctx, obj, PDF_NAME_None))
+		else if (pdf_is_name(ctx, obj) && pdf_name_eq(ctx, obj, PDF_NAME(None)))
 		{
 			proc->op_gs_SMask(ctx, proc, NULL, NULL, NULL, 0);
 		}
@@ -307,32 +311,30 @@ pdf_process_Do(fz_context *ctx, pdf_processor *proc, pdf_csi *csi)
 {
 	pdf_obj *xres, *xobj, *subtype;
 
-	xres = pdf_dict_get(ctx, csi->rdb, PDF_NAME_XObject);
-	if (!xres)
-		fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find XObject dictionary");
+	xres = pdf_dict_get(ctx, csi->rdb, PDF_NAME(XObject));
 	xobj = pdf_dict_gets(ctx, xres, csi->name);
 	if (!xobj)
-		fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find XObject resource '%s'", csi->name);
-	subtype = pdf_dict_get(ctx, xobj, PDF_NAME_Subtype);
-	if (pdf_name_eq(ctx, subtype, PDF_NAME_Form))
+		fz_throw(ctx, FZ_ERROR_MINOR, "cannot find XObject resource '%s'", csi->name);
+	subtype = pdf_dict_get(ctx, xobj, PDF_NAME(Subtype));
+	if (pdf_name_eq(ctx, subtype, PDF_NAME(Form)))
 	{
-		pdf_obj *st = pdf_dict_get(ctx, xobj, PDF_NAME_Subtype2);
+		pdf_obj *st = pdf_dict_get(ctx, xobj, PDF_NAME(Subtype2));
 		if (st)
 			subtype = st;
 	}
 	if (!pdf_is_name(ctx, subtype))
-		fz_throw(ctx, FZ_ERROR_SYNTAX, "no XObject subtype specified");
+		fz_throw(ctx, FZ_ERROR_MINOR, "no XObject subtype specified");
 
-	if (pdf_is_hidden_ocg(ctx, csi->doc->ocg, csi->rdb, proc->usage, pdf_dict_get(ctx, xobj, PDF_NAME_OC)))
+	if (pdf_is_hidden_ocg(ctx, csi->doc->ocg, csi->rdb, proc->usage, pdf_dict_get(ctx, xobj, PDF_NAME(OC))))
 		return;
 
-	if (pdf_name_eq(ctx, subtype, PDF_NAME_Form))
+	if (pdf_name_eq(ctx, subtype, PDF_NAME(Form)))
 	{
 		if (proc->op_Do_form)
 			proc->op_Do_form(ctx, proc, csi->name, xobj, csi->rdb);
 	}
 
-	else if (pdf_name_eq(ctx, subtype, PDF_NAME_Image))
+	else if (pdf_name_eq(ctx, subtype, PDF_NAME(Image)))
 	{
 		if (proc->op_Do_image)
 		{
@@ -378,12 +380,10 @@ pdf_process_CS(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, int stroke)
 		else
 		{
 			pdf_obj *csres, *csobj;
-			csres = pdf_dict_get(ctx, csi->rdb, PDF_NAME_ColorSpace);
-			if (!csres)
-				fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find ColorSpace dictionary");
+			csres = pdf_dict_get(ctx, csi->rdb, PDF_NAME(ColorSpace));
 			csobj = pdf_dict_gets(ctx, csres, csi->name);
 			if (!csobj)
-				fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find ColorSpace resource '%s'", csi->name);
+				fz_throw(ctx, FZ_ERROR_MINOR, "cannot find ColorSpace resource '%s'", csi->name);
 			cs = pdf_load_colorspace(ctx, csobj);
 		}
 
@@ -408,14 +408,12 @@ pdf_process_SC(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, int stroke)
 	{
 		pdf_obj *patres, *patobj, *type;
 
-		patres = pdf_dict_get(ctx, csi->rdb, PDF_NAME_Pattern);
-		if (!patres)
-			fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find Pattern dictionary");
+		patres = pdf_dict_get(ctx, csi->rdb, PDF_NAME(Pattern));
 		patobj = pdf_dict_gets(ctx, patres, csi->name);
 		if (!patobj)
-			fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find Pattern resource '%s'", csi->name);
+			fz_throw(ctx, FZ_ERROR_MINOR, "cannot find Pattern resource '%s'", csi->name);
 
-		type = pdf_dict_get(ctx, patobj, PDF_NAME_PatternType);
+		type = pdf_dict_get(ctx, patobj, PDF_NAME(PatternType));
 
 		if (pdf_to_int(ctx, type) == 1)
 		{
@@ -457,7 +455,7 @@ pdf_process_SC(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, int stroke)
 
 		else
 		{
-			fz_throw(ctx, FZ_ERROR_SYNTAX, "unknown pattern type: %d", pdf_to_int(ctx, type));
+			fz_throw(ctx, FZ_ERROR_MINOR, "unknown pattern type: %d", pdf_to_int(ctx, type));
 		}
 	}
 
@@ -477,7 +475,7 @@ static pdf_obj *
 resolve_properties(fz_context *ctx, pdf_csi *csi, pdf_obj *obj)
 {
 	if (pdf_is_name(ctx, obj))
-		return pdf_dict_get(ctx, pdf_dict_get(ctx, csi->rdb, PDF_NAME_Properties), obj);
+		return pdf_dict_get(ctx, pdf_dict_get(ctx, csi->rdb, PDF_NAME(Properties)), obj);
 	else
 		return obj;
 }
@@ -549,6 +547,18 @@ pdf_process_end(fz_context *ctx, pdf_processor *proc, pdf_csi *csi)
 		proc->op_END(ctx, proc);
 }
 
+static int is_known_bad_word(const char *word)
+{
+	switch (*word)
+	{
+	case 'I': return !strcmp(word, "Infinity");
+	case 'N': return !strcmp(word, "NaN");
+	case 'i': return !strcmp(word, "inf");
+	case 'n': return !strcmp(word, "nan");
+	}
+	return 0;
+}
+
 #define A(a) (a)
 #define B(a,b) (a | b << 8)
 #define C(a,b,c) (a | b << 8 | c << 16)
@@ -557,6 +567,7 @@ static void
 pdf_process_keyword(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_stream *stm, char *word)
 {
 	float *s = csi->stack;
+	char csname[40];
 	int key;
 
 	key = word[0];
@@ -575,7 +586,12 @@ pdf_process_keyword(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_strea
 	{
 	default:
 		if (!csi->xbalance)
-			fz_throw(ctx, FZ_ERROR_SYNTAX, "unknown keyword: '%s'", word);
+		{
+			if (is_known_bad_word(word))
+				fz_throw(ctx, FZ_ERROR_MINOR, "unknown keyword: '%s'", word);
+			else
+				fz_throw(ctx, FZ_ERROR_SYNTAX, "unknown keyword: '%s'", word);
+		}
 		break;
 
 	/* general graphics state */
@@ -590,12 +606,10 @@ pdf_process_keyword(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_strea
 	case B('g','s'):
 		{
 			pdf_obj *gsres, *gsobj;
-			gsres = pdf_dict_get(ctx, csi->rdb, PDF_NAME_ExtGState);
-			if (!gsres)
-				fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find ExtGState dictionary");
+			gsres = pdf_dict_get(ctx, csi->rdb, PDF_NAME(ExtGState));
 			gsobj = pdf_dict_gets(ctx, gsres, csi->name);
 			if (!gsobj)
-				fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find ExtGState resource '%s'", csi->name);
+				fz_throw(ctx, FZ_ERROR_MINOR, "cannot find ExtGState resource '%s'", csi->name);
 			if (proc->op_gs_begin)
 				proc->op_gs_begin(ctx, proc, csi->name, gsobj);
 			pdf_process_extgstate(ctx, proc, csi, gsobj);
@@ -651,13 +665,12 @@ pdf_process_keyword(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_strea
 		{
 			pdf_obj *fontres, *fontobj;
 			pdf_font_desc *font;
-			fontres = pdf_dict_get(ctx, csi->rdb, PDF_NAME_Font);
-			if (!fontres)
-				fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find Font dictionary");
+			fontres = pdf_dict_get(ctx, csi->rdb, PDF_NAME(Font));
 			fontobj = pdf_dict_gets(ctx, fontres, csi->name);
-			if (!fontobj)
-				fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find Font resource '%s'", csi->name);
-			font = load_font_or_hail_mary(ctx, csi->doc, csi->rdb, fontobj, 0, csi->cookie);
+			if (pdf_is_dict(ctx, fontobj))
+				font = pdf_try_load_font(ctx, csi->doc, csi->rdb, fontobj, csi->cookie);
+			else
+				font = pdf_load_hail_mary_font(ctx, csi->doc);
 			fz_try(ctx)
 				proc->op_Tf(ctx, proc, csi->name, font, s[0]);
 			fz_always(ctx)
@@ -725,11 +738,11 @@ pdf_process_keyword(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_strea
 	/* shadings, images, xobjects */
 	case B('B','I'):
 		{
-			fz_image *img = parse_inline_image(ctx, csi, stm);
+			fz_image *img = parse_inline_image(ctx, csi, stm, csname, sizeof csname);
 			fz_try(ctx)
 			{
 				if (proc->op_BI)
-					proc->op_BI(ctx, proc, img);
+					proc->op_BI(ctx, proc, img, csname[0] ? csname : NULL);
 			}
 			fz_always(ctx)
 				fz_drop_image(ctx, img);
@@ -743,12 +756,10 @@ pdf_process_keyword(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_strea
 		{
 			pdf_obj *shaderes, *shadeobj;
 			fz_shade *shade;
-			shaderes = pdf_dict_get(ctx, csi->rdb, PDF_NAME_Shading);
-			if (!shaderes)
-				fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find Shading dictionary");
+			shaderes = pdf_dict_get(ctx, csi->rdb, PDF_NAME(Shading));
 			shadeobj = pdf_dict_gets(ctx, shaderes, csi->name);
 			if (!shadeobj)
-				fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find Shading resource '%s'", csi->name);
+				fz_throw(ctx, FZ_ERROR_MINOR, "cannot find Shading resource '%s'", csi->name);
 			shade = pdf_load_shading(ctx, csi->doc, shadeobj);
 			fz_try(ctx)
 				proc->op_sh(ctx, proc, csi->name, shade);
@@ -892,14 +903,14 @@ pdf_process_stream(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_stream
 					{
 						pdf_drop_obj(ctx, csi->obj);
 						csi->obj = NULL;
-						csi->obj = pdf_new_name(ctx, doc, buf->scratch);
+						csi->obj = pdf_new_name(ctx, buf->scratch);
 					}
 					else
 						fz_strlcpy(csi->name, buf->scratch, sizeof(csi->name));
 					break;
 
 				case PDF_TOK_INT:
-					if (csi->top < nelem(csi->stack)) {
+					if (csi->top < (int)nelem(csi->stack)) {
 						csi->stack[csi->top] = buf->i;
 						csi->top ++;
 					}
@@ -908,7 +919,7 @@ pdf_process_stream(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_stream
 					break;
 
 				case PDF_TOK_REAL:
-					if (csi->top < nelem(csi->stack)) {
+					if (csi->top < (int)nelem(csi->stack)) {
 						csi->stack[csi->top] = buf->f;
 						csi->top ++;
 					}
@@ -929,7 +940,7 @@ pdf_process_stream(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_stream
 							pdf_drop_obj(ctx, csi->obj);
 							csi->obj = NULL;
 						}
-						csi->obj = pdf_new_string(ctx, doc, buf->scratch, buf->len);
+						csi->obj = pdf_new_string(ctx, buf->scratch, buf->len);
 					}
 					break;
 
@@ -951,20 +962,21 @@ pdf_process_stream(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_stream
 		fz_catch(ctx)
 		{
 			int caught = fz_caught(ctx);
-
 			if (cookie)
 			{
 				if (caught == FZ_ERROR_TRYLATER)
 				{
-					if (cookie->incomplete_ok)
-						cookie->incomplete++;
-					else
-						fz_rethrow(ctx);
+					cookie->incomplete++;
+					tok = PDF_TOK_EOF;
 				}
 				else if (caught == FZ_ERROR_ABORT)
 				{
 					fz_rethrow(ctx);
 				}
+				else if (caught == FZ_ERROR_MINOR)
+				{
+					cookie->errors++;
+				}
 				else if (caught == FZ_ERROR_SYNTAX)
 				{
 					cookie->errors++;
@@ -976,17 +988,17 @@ pdf_process_stream(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_stream
 				}
 				else
 				{
-					cookie->errors++;
-					fz_warn(ctx, "unrecoverable error; ignoring rest of page");
-					tok = PDF_TOK_EOF;
+					fz_rethrow(ctx);
 				}
 			}
 			else
 			{
 				if (caught == FZ_ERROR_TRYLATER)
-					fz_rethrow(ctx);
+					tok = PDF_TOK_EOF;
 				else if (caught == FZ_ERROR_ABORT)
 					fz_rethrow(ctx);
+				else if (caught == FZ_ERROR_MINOR)
+					/* ignore minor errors */ ;
 				else if (caught == FZ_ERROR_SYNTAX)
 				{
 					if (++syntax_errors >= MAX_SYNTAX_ERRORS)
@@ -997,8 +1009,7 @@ pdf_process_stream(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_stream
 				}
 				else
 				{
-					fz_warn(ctx, "unrecoverable error; ignoring rest of page");
-					tok = PDF_TOK_EOF;
+					fz_rethrow(ctx);
 				}
 			}
 
@@ -1040,14 +1051,29 @@ pdf_process_contents(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pd
 	}
 	fz_catch(ctx)
 	{
+		proc->close_processor = NULL; /* aborted run, don't warn about unclosed processor */
 		fz_rethrow(ctx);
 	}
 }
 
+/* Bug 702543: It looks like certain types of annotation are never
+ * printed. */
+static int
+pdf_should_print_annot(fz_context *ctx, pdf_annot *annot)
+{
+	enum pdf_annot_type type = pdf_annot_type(ctx, annot);
+
+	/* We may need to add more types here. */
+	if (type == PDF_ANNOT_FILE_ATTACHMENT)
+		return 0;
+
+	return 1;
+}
+
 void
 pdf_process_annot(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pdf_page *page, pdf_annot *annot, fz_cookie *cookie)
 {
-	int flags = pdf_to_int(ctx, pdf_dict_get(ctx, annot->obj, PDF_NAME_F));
+	int flags = pdf_dict_get_int(ctx, annot->obj, PDF_NAME(F));
 
 	if (flags & (PDF_ANNOT_IS_INVISIBLE | PDF_ANNOT_IS_HIDDEN))
 		return;
@@ -1058,8 +1084,13 @@ pdf_process_annot(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pdf_p
 
 	if (proc->usage)
 	{
-		if (!strcmp(proc->usage, "Print") && !(flags & PDF_ANNOT_IS_PRINT))
-			return;
+		if (!strcmp(proc->usage, "Print"))
+		{
+			if (!(flags & PDF_ANNOT_IS_PRINT))
+				return;
+			if (!pdf_should_print_annot(ctx, annot))
+				return;
+		}
 		if (!strcmp(proc->usage, "View") && (flags & PDF_ANNOT_IS_NO_VIEW))
 			return;
 	}
@@ -1067,13 +1098,12 @@ pdf_process_annot(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pdf_p
 	/* TODO: NoZoom and NoRotate */
 
 	/* XXX what resources, if any, to use for this check? */
-	if (pdf_is_hidden_ocg(ctx, doc->ocg, NULL, proc->usage, pdf_dict_get(ctx, annot->obj, PDF_NAME_OC)))
+	if (pdf_is_hidden_ocg(ctx, doc->ocg, NULL, proc->usage, pdf_dict_get(ctx, annot->obj, PDF_NAME(OC))))
 		return;
 
 	if (proc->op_q && proc->op_cm && proc->op_Do_form && proc->op_Q && annot->ap)
 	{
-		fz_matrix matrix;
-		pdf_annot_transform(ctx, annot, &matrix);
+		fz_matrix matrix = pdf_annot_transform(ctx, annot);
 		proc->op_q(ctx, proc);
 		proc->op_cm(ctx, proc,
 			matrix.a, matrix.b,
@@ -1113,6 +1143,11 @@ pdf_process_glyph(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pdf_o
 	}
 	fz_catch(ctx)
 	{
+		/* Note: Any SYNTAX errors should have been swallowed
+		 * by pdf_process_stream, but in case any escape from other
+		 * functions, recast the error type here to be safe. */
+		if (fz_caught(ctx) == FZ_ERROR_SYNTAX)
+			fz_throw(ctx, FZ_ERROR_GENERIC, "syntax error in content stream");
 		fz_rethrow(ctx);
 	}
 }
@@ -1179,14 +1214,14 @@ pdf_tos_make_trm(fz_context *ctx, pdf_text_object_state *tos, pdf_text_state *te
 		tos->char_ty = w1 * text->size + text->char_space;
 	}
 
-	fz_concat(trm, &tsm, &tos->tm);
+	*trm = fz_concat(tsm, tos->tm);
 
 	tos->cid = cid;
 	tos->gid = pdf_font_cid_to_gid(ctx, fontdesc, cid);
 	tos->fontdesc = fontdesc;
 
 	/* Compensate for the glyph cache limited positioning precision */
-	fz_expand_rect(fz_bound_glyph(ctx, fontdesc->font, tos->gid, trm, &tos->char_bbox), 1);
+	tos->char_bbox = fz_expand_rect(fz_bound_glyph(ctx, fontdesc->font, tos->gid, *trm), 1);
 
 	return tos->gid;
 }
@@ -1194,15 +1229,14 @@ pdf_tos_make_trm(fz_context *ctx, pdf_text_object_state *tos, pdf_text_state *te
 void
 pdf_tos_move_after_char(fz_context *ctx, pdf_text_object_state *tos)
 {
-	fz_union_rect(&tos->text_bbox, &tos->char_bbox);
-
-	fz_pre_translate(&tos->tm, tos->char_tx, tos->char_ty);
+	tos->text_bbox = fz_union_rect(tos->text_bbox, tos->char_bbox);
+	tos->tm = fz_pre_translate(tos->tm, tos->char_tx, tos->char_ty);
 }
 
 void
 pdf_tos_translate(pdf_text_object_state *tos, float tx, float ty)
 {
-	fz_pre_translate(&tos->tlm, tx, ty);
+	tos->tlm = fz_pre_translate(tos->tlm, tx, ty);
 	tos->tm = tos->tlm;
 }
 
@@ -1221,6 +1255,6 @@ pdf_tos_set_matrix(pdf_text_object_state *tos, float a, float b, float c, float
 void
 pdf_tos_newline(pdf_text_object_state *tos, float leading)
 {
-	fz_pre_translate(&tos->tlm, 0, -leading);
+	tos->tlm = fz_pre_translate(tos->tlm, 0, -leading);
 	tos->tm = tos->tlm;
 }
