@@ -1,5 +1,8 @@
+#include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
+#include <assert.h>
+
 static int
 pdf_make_hash_key(fz_context *ctx, fz_store_hash *hash, void *key_)
 {
@@ -31,23 +34,30 @@ pdf_cmp_key(fz_context *ctx, void *k0, void *k1)
 }
 
 static void
-pdf_print_key(fz_context *ctx, fz_output *out, void *key_)
+pdf_format_key(fz_context *ctx, char *s, size_t n, void *key_)
 {
 	pdf_obj *key = (pdf_obj *)key_;
-
 	if (pdf_is_indirect(ctx, key))
-		fz_write_printf(ctx, out, "(%d 0 R) ", pdf_to_num(ctx, key));
+		fz_snprintf(s, n, "(%d 0 R)", pdf_to_num(ctx, key));
 	else
-		pdf_print_obj(ctx, out, key, 0);
+	{
+		size_t t;
+		char *p = pdf_sprint_obj(ctx, s, n, &t, key, 1, 0);
+		if (p != s) {
+			fz_strlcpy(s, p, n);
+			fz_free(ctx, p);
+		}
+	}
 }
 
 static const fz_store_type pdf_obj_store_type =
 {
+	"pdf_obj",
 	pdf_make_hash_key,
 	pdf_keep_key,
 	pdf_drop_key,
 	pdf_cmp_key,
-	pdf_print_key,
+	pdf_format_key,
 	NULL
 };
 
