@@ -71,13 +71,13 @@
 	MURASTER_CONFIG_WIDTH: The printable page width
 	(in inches)
 */
-/* #define MURASTER_CONFIG_WIDTH 8.27 */
+/* #define MURASTER_CONFIG_WIDTH 8.27f */
 
 /*
 	MURASTER_CONFIG_HEIGHT: The printable page height
 	(in inches)
 */
-/* #define MURASTER_CONFIG_HEIGHT 11.69 */
+/* #define MURASTER_CONFIG_HEIGHT 11.69f */
 
 /*
 	MURASTER_CONFIG_STORE_SIZE: The maximum size to use
@@ -139,12 +139,39 @@
 #include "mupdf/fitz.h"
 #include "mupdf/helpers/mu-threads.h"
 
+#include <string.h>
+#include <stdlib.h>
+#include <stdio.h>
+
+#ifdef _MSC_VER
+struct timeval;
+struct timezone;
+int gettimeofday(struct timeval *tv, struct timezone *tz);
+#else
+#include <sys/time.h>
+#endif
+
 /*
 	After this point, we convert the #defines set (or not set)
 	above into sensible values we can work with. Don't edit
 	these for configuration.
 */
 
+/* Unless we have specifically disabled threading, enable it. */
+#ifndef DISABLE_MUTHREADS
+#ifndef MURASTER_THREADS
+#define MURASTER_THREADS 1
+#endif
+#endif
+
+/* If we have threading, and we haven't already configured BGPRINT,
+ * enable it. */
+#if MURASTER_THREADS != 0
+#ifndef MURASTER_CONFIG_BGPRINT
+#define MURASTER_CONFIG_BGPRINT 1
+#endif
+#endif
+
 #ifdef MURASTER_CONFIG_X_RESOLUTION
 #define X_RESOLUTION MURASTER_CONFIG_X_RESOLUTION
 #else
@@ -160,13 +187,13 @@
 #ifdef MURASTER_CONFIG_WIDTH
 #define PAPER_WIDTH MURASTER_CONFIG_WIDTH
 #else
-#define PAPER_WIDTH 8.27
+#define PAPER_WIDTH 8.27f
 #endif
 
 #ifdef MURASTER_CONFIG_HEIGHT
 #define PAPER_HEIGHT MURASTER_CONFIG_HEIGHT
 #else
-#define PAPER_HEIGHT 11.69
+#define PAPER_HEIGHT 11.69f
 #endif
 
 #ifdef MURASTER_CONFIG_STORE_SIZE
@@ -354,16 +381,13 @@ static int width = 0;
 static int height = 0;
 static int fit = 0;
 
-static float layout_w = 450;
-static float layout_h = 600;
-static float layout_em = 12;
+static float layout_w = FZ_DEFAULT_LAYOUT_W;
+static float layout_h = FZ_DEFAULT_LAYOUT_H;
+static float layout_em = FZ_DEFAULT_LAYOUT_EM;
 static char *layout_css = NULL;
 static int layout_use_doc_css = 1;
 
 static int showtime = 0;
-static size_t memtrace_current = 0;
-static size_t memtrace_peak = 0;
-static size_t memtrace_total = 0;
 static int showmemory = 0;
 
 static int ignore_errors = 0;
@@ -372,7 +396,6 @@ static int alphabits_graphics = 8;
 
 static int min_band_height;
 static size_t max_band_memory;
-int band_height;
 
 static int errored = 0;
 static fz_colorspace *colorspace;
@@ -514,7 +537,7 @@ static int gettime(void)
 	return (now.tv_sec - first.tv_sec) * 1000 + (now.tv_usec - first.tv_usec) / 1000;
 }
 
-static int drawband(fz_context *ctx, fz_page *page, fz_display_list *list, const fz_matrix *ctm, const fz_rect *tbounds, fz_cookie *cookie, int band_start, fz_pixmap *pix, fz_bitmap **bit)
+static int drawband(fz_context *ctx, fz_page *page, fz_display_list *list, fz_matrix ctm, fz_rect tbounds, fz_cookie *cookie, int band_start, fz_pixmap *pix, fz_bitmap **bit)
 {
 	fz_device *dev = NULL;
 
@@ -524,7 +547,7 @@ static int drawband(fz_context *ctx, fz_page *page, fz_display_list *list, const
 	{
 		fz_clear_pixmap_with_value(ctx, pix, 255);
 
-		dev = fz_new_draw_device(ctx, NULL, pix);
+		dev = fz_new_draw_device(ctx, fz_identity, pix);
 		if (alphabits_graphics == 0)
 			fz_enable_device_hints(ctx, dev, FZ_DONT_INTERPOLATE_IMAGES);
 		if (list)
@@ -551,6 +574,16 @@ static int dodrawpage(fz_context *ctx, int pagenum, fz_cookie *cookie, render_de
 	fz_pixmap *pix = NULL;
 	fz_bitmap *bit = NULL;
 	int errors_are_fatal = 0;
+	fz_irect ibounds = render->ibounds;
+	fz_rect tbounds = render->tbounds;
+	int total_height = ibounds.y1 - ibounds.y0;
+	int start_offset = min_band_height * render->bands_rendered;
+	int remaining_start = ibounds.y0 + start_offset;
+	int remaining_height = ibounds.y1 - remaining_start;
+	int band_height = min_band_height * render->band_height_multiple;
+	int bands = (remaining_height + band_height-1) / band_height;
+	fz_matrix ctm = render->ctm;
+	int band;
 
 	fz_var(pix);
 	fz_var(bit);
@@ -558,17 +591,6 @@ static int dodrawpage(fz_context *ctx, int pagenum, fz_cookie *cookie, render_de
 
 	fz_try(ctx)
 	{
-		fz_irect ibounds = render->ibounds;
-		fz_rect tbounds = render->tbounds;
-		int total_height = ibounds.y1 - ibounds.y0;
-		int start_offset = min_band_height * render->bands_rendered;
-		int remaining_start = ibounds.y0 + start_offset;
-		int remaining_height = ibounds.y1 - remaining_start;
-		int band_height = min_band_height * render->band_height_multiple;
-		int bands = (remaining_height + band_height-1) / band_height;
-		fz_matrix ctm = render->ctm;
-		int band;
-
 		/* Set up ibounds and tbounds for a single band_height band.
 		 * We will adjust ctm as we go. */
 		ibounds.y1 = ibounds.y0 + band_height;
@@ -590,7 +612,7 @@ static int dodrawpage(fz_context *ctx, int pagenum, fz_cookie *cookie, render_de
 				if (remaining_height < band_height)
 					ibounds.y1 = ibounds.y0 + remaining_height;
 				remaining_height -= band_height;
-				w->pix = fz_new_pixmap_with_bbox(ctx, colorspace, &ibounds, 0);
+				w->pix = fz_new_pixmap_with_bbox(ctx, colorspace, ibounds, NULL, 0);
 				fz_set_pixmap_resolution(ctx, w->pix, x_resolution, y_resolution);
 				DEBUG_THREADS(("Worker %d, Pre-triggering band %d\n", band, band));
 				w->started = 1;
@@ -601,10 +623,9 @@ static int dodrawpage(fz_context *ctx, int pagenum, fz_cookie *cookie, render_de
 		}
 		else
 		{
-			pix = fz_new_pixmap_with_bbox(ctx, colorspace, &ibounds, 0);
+			pix = fz_new_pixmap_with_bbox(ctx, colorspace, ibounds, NULL, 0);
 			fz_set_pixmap_resolution(ctx, pix, x_resolution, y_resolution);
 		}
-		fz_write_header(ctx, render->bander, pix->w, total_height, pix->n, pix->alpha, pix->xres, pix->yres, pagenum);
 
 		for (band = 0; band < bands; band++)
 		{
@@ -628,7 +649,7 @@ static int dodrawpage(fz_context *ctx, int pagenum, fz_cookie *cookie, render_de
 				cookie->errors += w->cookie.errors;
 			}
 			else
-				status = drawband(ctx, render->page, render->list, &ctm, &tbounds, cookie, band_start, pix, &bit);
+				status = drawband(ctx, render->page, render->list, ctm, tbounds, cookie, band_start, pix, &bit);
 
 			if (status != RENDER_OK)
 				fz_throw(ctx, FZ_ERROR_GENERIC, "Render failed");
@@ -666,7 +687,7 @@ static int dodrawpage(fz_context *ctx, int pagenum, fz_cookie *cookie, render_de
 		if (render->num_workers > 0)
 		{
 			int band;
-			for (band = 0; band < render->num_workers; band++)
+			for (band = 0; band < fz_mini(render->num_workers, bands); band++)
 			{
 				worker_t *w = &workers[band];
 				w->cookie.abort = 1;
@@ -695,7 +716,7 @@ static int dodrawpage(fz_context *ctx, int pagenum, fz_cookie *cookie, render_de
 }
 
 /* This functions tries to render a page, falling back repeatedly to try and make it work. */
-static int try_render_page(fz_context *ctx, int pagenum, fz_cookie *cookie, int start, int interptime, char *filename, int bg, int solo, render_details *render)
+static int try_render_page(fz_context *ctx, int pagenum, fz_cookie *cookie, int start, int interptime, char *fname, int bg, int solo, render_details *render)
 {
 	int status;
 
@@ -707,7 +728,7 @@ static int try_render_page(fz_context *ctx, int pagenum, fz_cookie *cookie, int
 		{
 			int w = render->ibounds.x1 - render->ibounds.x0;
 			int h = render->ibounds.y1 - render->ibounds.y0;
-			fz_write_header(ctx, render->bander, w, h, render->n, 0, 0, 0, 0);
+			fz_write_header(ctx, render->bander, w, h, render->n, 0, 0, 0, 0, 0, NULL);
 		}
 		fz_catch(ctx)
 		{
@@ -775,14 +796,14 @@ static int try_render_page(fz_context *ctx, int pagenum, fz_cookie *cookie, int
 				timing.min = diff + interptime;
 				timing.mininterp = interptime;
 				timing.minpage = pagenum;
-				timing.minfilename = filename;
+				timing.minfilename = fname;
 			}
 			if (diff + interptime > timing.max)
 			{
 				timing.max = diff + interptime;
 				timing.maxinterp = interptime;
 				timing.maxpage = pagenum;
-				timing.maxfilename = filename;
+				timing.maxfilename = fname;
 			}
 			timing.total += diff + interptime;
 			timing.count ++;
@@ -795,13 +816,13 @@ static int try_render_page(fz_context *ctx, int pagenum, fz_cookie *cookie, int
 			{
 				timing.min = diff;
 				timing.minpage = pagenum;
-				timing.minfilename = filename;
+				timing.minfilename = fname;
 			}
 			if (diff > timing.max)
 			{
 				timing.max = diff;
 				timing.maxpage = pagenum;
-				timing.maxfilename = filename;
+				timing.maxfilename = fname;
 			}
 			timing.total += diff;
 			timing.count ++;
@@ -812,7 +833,7 @@ static int try_render_page(fz_context *ctx, int pagenum, fz_cookie *cookie, int
 
 	if (showmemory)
 	{
-		fz_dump_glyph_cache_stats(ctx);
+		fz_dump_glyph_cache_stats(ctx, fz_stderr(ctx));
 	}
 
 	fz_flush_warnings(ctx);
@@ -841,7 +862,7 @@ get_page_render_details(fz_context *ctx, fz_page *page, render_details *render)
 	render->list = NULL;
 	render->num_workers = num_workers;
 
-	fz_bound_page(ctx, page, &render->bounds);
+	render->bounds = fz_bound_page(ctx, page);
 	page_width = (render->bounds.x1 - render->bounds.x0)/72;
 	page_height = (render->bounds.y1 - render->bounds.y0)/72;
 
@@ -908,9 +929,9 @@ get_page_render_details(fz_context *ctx, fz_page *page, render_details *render)
 		rot = rotation;
 	}
 
-	fz_pre_scale(fz_rotate(&render->ctm, rot), s_x, s_y);
-	render->tbounds = render->bounds;
-	fz_round_rect(&render->ibounds, fz_transform_rect(&render->tbounds, &render->ctm));
+	render->ctm = fz_pre_scale(fz_rotate(rot), s_x, s_y);
+	render->tbounds = fz_transform_rect(render->bounds, render->ctm);;
+	render->ibounds = fz_round_rect(render->tbounds);
 }
 
 static void
@@ -956,7 +977,7 @@ initialise_banding(fz_context *ctx, render_details *render, int color)
 	}
 
 	w = render->ibounds.x1 - render->ibounds.x0;
-	min_band_mem = bpp * w * min_band_height;
+	min_band_mem = (size_t)bpp * w * min_band_height;
 	reps = (int)(max_band_memory / min_band_mem);
 	if (reps < 1)
 		reps = 1;
@@ -981,7 +1002,7 @@ initialise_banding(fz_context *ctx, render_details *render, int color)
 	if (output_format == OUT_PGM || output_format == OUT_PPM)
 	{
 		render->bander = fz_new_pnm_band_writer(ctx, out);
-		render->n = OUT_PGM ? 1 : 3;
+		render->n = output_format == OUT_PGM ? 1 : 3;
 	}
 	else if (output_format == OUT_PAM)
 	{
@@ -1032,14 +1053,14 @@ static void drawpage(fz_context *ctx, fz_document *doc, int pagenum)
 		/* Make the display list, and see if we need color */
 		fz_try(ctx)
 		{
-			list = fz_new_display_list(ctx, &render.bounds);
+			list = fz_new_display_list(ctx, render.bounds);
 			list_dev = fz_new_list_device(ctx, list);
 #if GREY_FALLBACK != 0
 			test_dev = fz_new_test_device(ctx, &is_color, 0.01f, 0, list_dev);
-			fz_run_page(ctx, page, test_dev, &fz_identity, &cookie);
+			fz_run_page(ctx, page, test_dev, fz_identity, &cookie);
 			fz_close_device(ctx, test_dev);
 #else
-			fz_run_page(ctx, page, list_dev, &fz_identity, &cookie);
+			fz_run_page(ctx, page, list_dev, fz_identity, &cookie);
 #endif
 			fz_close_device(ctx, list_dev);
 		}
@@ -1068,8 +1089,8 @@ static void drawpage(fz_context *ctx, fz_document *doc, int pagenum)
 			 * from file. */
 			fz_try(ctx)
 			{
-				test_dev = fz_new_test_device(ctx, &is_color, 0.01f, 0, test_dev);
-				fz_run_page(ctx, page, test_dev, &fz_identity, &cookie);
+				test_dev = fz_new_test_device(ctx, &is_color, 0.01f, 0, NULL);
+				fz_run_page(ctx, page, test_dev, fz_identity, &cookie);
 				fz_close_device(ctx, test_dev);
 			}
 			fz_always(ctx)
@@ -1244,9 +1265,17 @@ typedef struct
 #endif
 } trace_header;
 
+typedef struct
+{
+	size_t current;
+	size_t peak;
+	size_t total;
+} trace_info;
+
 static void *
 trace_malloc(void *arg, size_t size)
 {
+	trace_info *info = (trace_info *) arg;
 	trace_header *p;
 	if (size == 0)
 		return NULL;
@@ -1254,27 +1283,29 @@ trace_malloc(void *arg, size_t size)
 	if (p == NULL)
 		return NULL;
 	p[0].size = size;
-	memtrace_current += size;
-	memtrace_total += size;
-	if (memtrace_current > memtrace_peak)
-		memtrace_peak = memtrace_current;
+	info->current += size;
+	info->total += size;
+	if (info->current > info->peak)
+		info->peak = info->current;
 	return (void *)&p[1];
 }
 
 static void
 trace_free(void *arg, void *p_)
 {
+	trace_info *info = (trace_info *) arg;
 	trace_header *p = (trace_header *)p_;
 
 	if (p == NULL)
 		return;
-	memtrace_current -= p[-1].size;
+	info->current -= p[-1].size;
 	free(&p[-1]);
 }
 
 static void *
 trace_realloc(void *arg, void *p_, size_t size)
 {
+	trace_info *info = (trace_info *) arg;
 	trace_header *p = (trace_header *)p_;
 	size_t oldsize;
 
@@ -1289,11 +1320,11 @@ trace_realloc(void *arg, void *p_, size_t size)
 	p = realloc(&p[-1], size + sizeof(trace_header));
 	if (p == NULL)
 		return NULL;
-	memtrace_current += size - oldsize;
+	info->current += size - oldsize;
 	if (size > oldsize)
-		memtrace_total += size - oldsize;
-	if (memtrace_current > memtrace_peak)
-		memtrace_peak = memtrace_current;
+		info->total += size - oldsize;
+	if (info->current > info->peak)
+		info->peak = info->current;
 	p[0].size = size;
 	return &p[1];
 }
@@ -1302,19 +1333,21 @@ trace_realloc(void *arg, void *p_, size_t size)
 static void worker_thread(void *arg)
 {
 	worker_t *me = (worker_t *)arg;
+	int band_start;
 
 	do
 	{
 		DEBUG_THREADS(("Worker %d waiting\n", me->num));
 		mu_wait_semaphore(&me->start);
+		band_start = me->band_start;
 		DEBUG_THREADS(("Worker %d woken for band_start %d\n", me->num, me->band_start));
 		me->status = RENDER_OK;
-		if (me->band_start >= 0)
-			me->status = drawband(me->ctx, NULL, me->list, &me->ctm, &me->tbounds, &me->cookie, me->band_start, me->pix, &me->bit);
-		DEBUG_THREADS(("Worker %d completed band_start %d (status=%d)\n", me->num, me->band_start, me->status));
+		if (band_start >= 0)
+			me->status = drawband(me->ctx, NULL, me->list, me->ctm, me->tbounds, &me->cookie, band_start, me->pix, &me->bit);
+		DEBUG_THREADS(("Worker %d completed band_start %d (status=%d)\n", me->num, band_start, me->status));
 		mu_trigger_semaphore(&me->stop);
 	}
-	while (me->band_start >= 0);
+	while (band_start >= 0);
 }
 
 static void bgprint_worker(void *arg)
@@ -1390,7 +1423,8 @@ int main(int argc, char **argv)
 	fz_document *doc = NULL;
 	int c;
 	fz_context *ctx;
-	fz_alloc_context alloc_ctx = { NULL, trace_malloc, trace_realloc, trace_free };
+	trace_info info = { 0, 0, 0 };
+	fz_alloc_context alloc_ctx = { &info, trace_malloc, trace_realloc, trace_free };
 	fz_locks_context *locks = NULL;
 
 	fz_var(doc);
@@ -1550,7 +1584,7 @@ int main(int argc, char **argv)
 	{
 		int i;
 
-		for (i = 0; i < nelem(suffix_table); i++)
+		for (i = 0; i < (int)nelem(suffix_table); i++)
 		{
 			if (!strcmp(format, suffix_table[i].suffix+1))
 			{
@@ -1559,7 +1593,7 @@ int main(int argc, char **argv)
 				break;
 			}
 		}
-		if (i == nelem(suffix_table))
+		if (i == (int)nelem(suffix_table))
 		{
 			fprintf(stderr, "Unknown output format '%s'\n", format);
 			exit(1);
@@ -1570,7 +1604,7 @@ int main(int argc, char **argv)
 		char *suffix = output;
 		int i;
 
-		for (i = 0; i < nelem(suffix_table); i++)
+		for (i = 0; i < (int)nelem(suffix_table); i++)
 		{
 			char *s = strstr(suffix, suffix_table[i].suffix);
 
@@ -1699,6 +1733,7 @@ int main(int argc, char **argv)
 	}
 #endif /* DISABLE_MUTHREADS */
 
+	fz_close_output(ctx, out);
 	fz_drop_output(ctx, out);
 	out = NULL;
 
@@ -1709,9 +1744,9 @@ int main(int argc, char **argv)
 
 	if (showmemory)
 	{
-		fprintf(stderr, "Total memory use = " FMT_zu " bytes\n", memtrace_total);
-		fprintf(stderr, "Peak memory use = " FMT_zu " bytes\n", memtrace_peak);
-		fprintf(stderr, "Current memory use = " FMT_zu " bytes\n", memtrace_current);
+		char buf[100];
+		fz_snprintf(buf, sizeof buf, "Memory use total=%zu peak=%zu current=%zu", info.total, info.peak, info.current);
+		fprintf(stderr, "%s\n", buf);
 	}
 
 	return (errored != 0);
