@@ -14,8 +14,18 @@
 #define nelem(a) (int)(sizeof (a) / sizeof (a)[0])
 
 #define REPINF 255
-#define MAXSUB REG_MAXSUB
-#define MAXPROG (32 << 10)
+#ifndef REG_MAXPROG
+#define REG_MAXPROG (32 << 10)
+#endif
+#ifndef REG_MAXREC
+#define REG_MAXREC 1024
+#endif
+#ifndef REG_MAXSPAN
+#define REG_MAXSPAN 64
+#endif
+#ifndef REG_MAXCLASS
+#define REG_MAXCLASS 16
+#endif
 
 typedef struct Reclass Reclass;
 typedef struct Renode Renode;
@@ -24,14 +34,14 @@ typedef struct Rethread Rethread;
 
 struct Reclass {
 	Rune *end;
-	Rune spans[64];
+	Rune spans[REG_MAXSPAN];
 };
 
 struct Reprog {
 	Reinst *start, *end;
 	int flags;
 	int nsub;
-	Reclass cclass[16];
+	Reclass cclass[REG_MAXCLASS];
 };
 
 struct cstate {
@@ -41,7 +51,7 @@ struct cstate {
 	const char *source;
 	int ncclass;
 	int nsub;
-	Renode *sub[MAXSUB];
+	Renode *sub[REG_MAXSUB];
 
 	int lookahead;
 	Rune yychar;
@@ -106,20 +116,29 @@ static int isunicodeletter(int c)
 
 static int nextrune(struct cstate *g)
 {
+	if (!*g->source) {
+		g->yychar = EOF;
+		return 0;
+	}
 	g->source += chartorune(&g->yychar, g->source);
 	if (g->yychar == '\\') {
+		if (!*g->source)
+			die(g, "unterminated escape sequence");
 		g->source += chartorune(&g->yychar, g->source);
 		switch (g->yychar) {
-		case 0: die(g, "unterminated escape sequence");
 		case 'f': g->yychar = '\f'; return 0;
 		case 'n': g->yychar = '\n'; return 0;
 		case 'r': g->yychar = '\r'; return 0;
 		case 't': g->yychar = '\t'; return 0;
 		case 'v': g->yychar = '\v'; return 0;
 		case 'c':
+			if (!g->source[0])
+				die(g, "unterminated escape sequence");
 			g->yychar = (*g->source++) & 31;
 			return 0;
 		case 'x':
+			if (!g->source[0] || !g->source[1])
+				die(g, "unterminated escape sequence");
 			g->yychar = hex(g, *g->source++) << 4;
 			g->yychar += hex(g, *g->source++);
 			if (g->yychar == 0) {
@@ -128,6 +147,8 @@ static int nextrune(struct cstate *g)
 			}
 			return 0;
 		case 'u':
+			if (!g->source[0] || !g->source[1] || !g->source[2] || !g->source[3])
+				die(g, "unterminated escape sequence");
 			g->yychar = hex(g, *g->source++) << 12;
 			g->yychar += hex(g, *g->source++) << 8;
 			g->yychar += hex(g, *g->source++) << 4;
@@ -137,6 +158,9 @@ static int nextrune(struct cstate *g)
 				return 1;
 			}
 			return 0;
+		case 0:
+			g->yychar = '0';
+			return 1;
 		}
 		if (strchr(ESCAPES, g->yychar))
 			return 1;
@@ -156,9 +180,9 @@ static int lexcount(struct cstate *g)
 	while (g->yychar != ',' && g->yychar != '}') {
 		g->yymin = g->yymin * 10 + dec(g, g->yychar);
 		g->yychar = *g->source++;
+		if (g->yymin >= REPINF)
+			die(g, "numeric overflow");
 	}
-	if (g->yymin < 0 || g->yymin >= REPINF)
-		die(g, "numeric overflow");
 
 	if (g->yychar == ',') {
 		g->yychar = *g->source++;
@@ -170,9 +194,9 @@ static int lexcount(struct cstate *g)
 			while (g->yychar != '}') {
 				g->yymax = g->yymax * 10 + dec(g, g->yychar);
 				g->yychar = *g->source++;
+				if (g->yymax >= REPINF)
+					die(g, "numeric overflow");
 			}
-			if (g->yymax < 0 || g->yymax >= REPINF)
-				die(g, "numeric overflow");
 		}
 	} else {
 		g->yymax = g->yymin;
@@ -193,7 +217,7 @@ static void addrange(struct cstate *g, Rune a, Rune b)
 {
 	if (a > b)
 		die(g, "invalid character class range");
-	if (g->yycc->end + 2 == g->yycc->spans + nelem(g->yycc->spans))
+	if (g->yycc->end + 2 >= g->yycc->spans + nelem(g->yycc->spans))
 		die(g, "too many character class ranges");
 	*g->yycc->end++ = a;
 	*g->yycc->end++ = b;
@@ -262,7 +286,7 @@ static int lexclass(struct cstate *g)
 
 	havesave = havedash = 0;
 	for (;;) {
-		if (g->yychar == 0)
+		if (g->yychar == EOF)
 			die(g, "unterminated character class");
 		if (!quoted && g->yychar == ']')
 			break;
@@ -353,7 +377,7 @@ static int lex(struct cstate *g)
 	}
 
 	switch (g->yychar) {
-	case 0:
+	case EOF:
 	case '$': case ')': case '*': case '+':
 	case '.': case '?': case '^': case '|':
 		return g->yychar;
@@ -481,7 +505,7 @@ static Renode *parseatom(struct cstate *g)
 	}
 	if (g->lookahead == L_REF) {
 		atom = newnode(g, P_REF);
-		if (g->yychar == 0 || g->yychar > g->nsub || !g->sub[g->yychar])
+		if (g->yychar == 0 || g->yychar >= g->nsub || !g->sub[g->yychar])
 			die(g, "invalid back-reference");
 		atom->n = g->yychar;
 		atom->x = g->sub[g->yychar];
@@ -492,7 +516,7 @@ static Renode *parseatom(struct cstate *g)
 		return newnode(g, P_ANY);
 	if (accept(g, '(')) {
 		atom = newnode(g, P_PAR);
-		if (g->nsub == MAXSUB)
+		if (g->nsub == REG_MAXSUB)
 			die(g, "too many captures");
 		atom->n = g->nsub++;
 		atom->x = parsealt(g);
@@ -550,16 +574,19 @@ static Renode *parserep(struct cstate *g)
 
 static Renode *parsecat(struct cstate *g)
 {
-	Renode *cat, *x;
-	if (g->lookahead && g->lookahead != '|' && g->lookahead != ')') {
-		cat = parserep(g);
-		while (g->lookahead && g->lookahead != '|' && g->lookahead != ')') {
-			x = cat;
+	Renode *cat, *head, **tail;
+	if (g->lookahead != EOF && g->lookahead != '|' && g->lookahead != ')') {
+		/* Build a right-leaning tree by splicing in new 'cat' at the tail. */
+		head = parserep(g);
+		tail = &head;
+		while (g->lookahead != EOF && g->lookahead != '|' && g->lookahead != ')') {
 			cat = newnode(g, P_CAT);
-			cat->x = x;
+			cat->x = *tail;
 			cat->y = parserep(g);
+			*tail = cat;
+			tail = &cat->y;
 		}
-		return cat;
+		return head;
 	}
 	return NULL;
 }
@@ -609,7 +636,7 @@ static int count(struct cstate *g, Renode *node)
 		if (min == max) n = count(g, node->x) * min;
 		else if (max < REPINF) n = count(g, node->x) * max + (max - min);
 		else n = count(g, node->x) * (min + 1) + 2;
-		if (n < 0 || n > MAXPROG) die(g, "program too large");
+		if (n < 0 || n > REG_MAXPROG) die(g, "program too large");
 		return n;
 	case P_PAR: return count(g, node->x) + 2;
 	case P_PLA: return count(g, node->x) + 2;
@@ -633,14 +660,15 @@ static void compile(Reprog *prog, Renode *node)
 	Reinst *inst, *split, *jump;
 	int i;
 
+loop:
 	if (!node)
 		return;
 
 	switch (node->type) {
 	case P_CAT:
 		compile(prog, node->x);
-		compile(prog, node->y);
-		break;
+		node = node->y;
+		goto loop;
 
 	case P_ALT:
 		split = emit(prog, I_SPLIT);
@@ -831,14 +859,19 @@ Reprog *regcompx(void *(*alloc)(void *ctx, void *p, int n), void *ctx,
 	g.prog = alloc(ctx, NULL, sizeof (Reprog));
 	if (!g.prog)
 		die(&g, "cannot allocate regular expression");
-	g.pstart = g.pend = alloc(ctx, NULL, sizeof (Renode) * strlen(pattern) * 2);
-	if (!g.pstart)
-		die(&g, "cannot allocate regular expression parse list");
+	n = strlen(pattern) * 2;
+	if (n > REG_MAXPROG)
+		die(&g, "program too large");
+	if (n > 0) {
+		g.pstart = g.pend = alloc(ctx, NULL, sizeof (Renode) * n);
+		if (!g.pstart)
+			die(&g, "cannot allocate regular expression parse list");
+	}
 
 	g.source = pattern;
 	g.ncclass = 0;
 	g.nsub = 1;
-	for (i = 0; i < MAXSUB; ++i)
+	for (i = 0; i < REG_MAXSUB; ++i)
 		g.sub[i] = 0;
 
 	g.prog->flags = cflags;
@@ -847,7 +880,7 @@ Reprog *regcompx(void *(*alloc)(void *ctx, void *p, int n), void *ctx,
 	node = parsealt(&g);
 	if (g.lookahead == ')')
 		die(&g, "unmatched ')'");
-	if (g.lookahead != 0)
+	if (g.lookahead != EOF)
 		die(&g, "syntax error");
 
 #ifdef TEST
@@ -856,7 +889,7 @@ Reprog *regcompx(void *(*alloc)(void *ctx, void *p, int n), void *ctx,
 #endif
 
 	n = 6 + count(&g, node);
-	if (n < 0 || n > MAXPROG)
+	if (n < 0 || n > REG_MAXPROG)
 		die(&g, "program too large");
 
 	g.prog->nsub = g.nsub;
@@ -958,87 +991,96 @@ static int strncmpcanon(const char *a, const char *b, int n)
 	return 0;
 }
 
-static int match(Reinst *pc, const char *sp, const char *bol, int flags, Resub *out)
+static int match(Reinst *pc, const char *sp, const char *bol, int flags, Resub *out, int depth)
 {
 	Resub scratch;
+	int result;
 	int i;
 	Rune c;
 
+	/* stack overflow */
+	if (depth > REG_MAXREC)
+		return -1;
+
 	for (;;) {
 		switch (pc->opcode) {
 		case I_END:
-			return 1;
+			return 0;
 		case I_JUMP:
 			pc = pc->x;
 			break;
 		case I_SPLIT:
 			scratch = *out;
-			if (match(pc->x, sp, bol, flags, &scratch)) {
+			result = match(pc->x, sp, bol, flags, &scratch, depth+1);
+			if (result == -1)
+				return -1;
+			if (result == 0) {
 				*out = scratch;
-				return 1;
+				return 0;
 			}
 			pc = pc->y;
 			break;
 
 		case I_PLA:
-			if (!match(pc->x, sp, bol, flags, out))
-				return 0;
+			result = match(pc->x, sp, bol, flags, out, depth+1);
+			if (result == -1)
+				return -1;
+			if (result == 1)
+				return 1;
 			pc = pc->y;
 			break;
 		case I_NLA:
 			scratch = *out;
-			if (match(pc->x, sp, bol, flags, &scratch))
-				return 0;
+			result = match(pc->x, sp, bol, flags, &scratch, depth+1);
+			if (result == -1)
+				return -1;
+			if (result == 0)
+				return 1;
 			pc = pc->y;
 			break;
 
 		case I_ANYNL:
+			if (!*sp) return 1;
 			sp += chartorune(&c, sp);
-			if (c == 0)
-				return 0;
 			pc = pc + 1;
 			break;
 		case I_ANY:
+			if (!*sp) return 1;
 			sp += chartorune(&c, sp);
-			if (c == 0)
-				return 0;
 			if (isnewline(c))
-				return 0;
+				return 1;
 			pc = pc + 1;
 			break;
 		case I_CHAR:
+			if (!*sp) return 1;
 			sp += chartorune(&c, sp);
-			if (c == 0)
-				return 0;
 			if (flags & REG_ICASE)
 				c = canon(c);
 			if (c != pc->c)
-				return 0;
+				return 1;
 			pc = pc + 1;
 			break;
 		case I_CCLASS:
+			if (!*sp) return 1;
 			sp += chartorune(&c, sp);
-			if (c == 0)
-				return 0;
 			if (flags & REG_ICASE) {
 				if (!incclasscanon(pc->cc, canon(c)))
-					return 0;
+					return 1;
 			} else {
 				if (!incclass(pc->cc, c))
-					return 0;
+					return 1;
 			}
 			pc = pc + 1;
 			break;
 		case I_NCCLASS:
+			if (!*sp) return 1;
 			sp += chartorune(&c, sp);
-			if (c == 0)
-				return 0;
 			if (flags & REG_ICASE) {
 				if (incclasscanon(pc->cc, canon(c)))
-					return 0;
+					return 1;
 			} else {
 				if (incclass(pc->cc, c))
-					return 0;
+					return 1;
 			}
 			pc = pc + 1;
 			break;
@@ -1046,10 +1088,10 @@ static int match(Reinst *pc, const char *sp, const char *bol, int flags, Resub *
 			i = out->sub[pc->n].ep - out->sub[pc->n].sp;
 			if (flags & REG_ICASE) {
 				if (strncmpcanon(sp, out->sub[pc->n].sp, i))
-					return 0;
+					return 1;
 			} else {
 				if (strncmp(sp, out->sub[pc->n].sp, i))
-					return 0;
+					return 1;
 			}
 			if (i > 0)
 				sp += i;
@@ -1067,7 +1109,7 @@ static int match(Reinst *pc, const char *sp, const char *bol, int flags, Resub *
 					break;
 				}
 			}
-			return 0;
+			return 1;
 		case I_EOL:
 			if (*sp == 0) {
 				pc = pc + 1;
@@ -1079,19 +1121,19 @@ static int match(Reinst *pc, const char *sp, const char *bol, int flags, Resub *
 					break;
 				}
 			}
-			return 0;
+			return 1;
 		case I_WORD:
 			i = sp > bol && iswordchar(sp[-1]);
 			i ^= iswordchar(sp[0]);
 			if (!i)
-				return 0;
+				return 1;
 			pc = pc + 1;
 			break;
 		case I_NWORD:
 			i = sp > bol && iswordchar(sp[-1]);
 			i ^= iswordchar(sp[0]);
 			if (i)
-				return 0;
+				return 1;
 			pc = pc + 1;
 			break;
 
@@ -1104,7 +1146,7 @@ static int match(Reinst *pc, const char *sp, const char *bol, int flags, Resub *
 			pc = pc + 1;
 			break;
 		default:
-			return 0;
+			return 1;
 		}
 	}
 }
@@ -1118,10 +1160,10 @@ int regexec(Reprog *prog, const char *sp, Resub *sub, int eflags)
 		sub = &scratch;
 
 	sub->nsub = prog->nsub;
-	for (i = 0; i < MAXSUB; ++i)
+	for (i = 0; i < REG_MAXSUB; ++i)
 		sub->sub[i].sp = sub->sub[i].ep = NULL;
 
-	return !match(prog->start, sp, sp, prog->flags | eflags, sub);
+	return match(prog->start, sp, sp, prog->flags | eflags, sub, 0);
 }
 
 #ifdef TEST
